"""Utilities for processing BaseSpace autoimports."""


def get_line(basespace_autoimport):
    """Build a list of relevant data to be printed.

    Args:
        basespace_autoimport (BaseSpaceProjectImport): instance of
            BaseSpace autoimport

    Returns:
        list(str): list of relevant data to be printed
    """
    return "\t".join(
        [
            str(basespace_autoimport.id),
            str(basespace_autoimport.project_id),
            str(basespace_autoimport.identifier),
        ]
    )



"""VCR filters for samples tests."""
from gencove.tests.filters import _replace_uuid_from_url


def filter_samples_request(request):
    """Filter samples sensitive data from request."""
    return _replace_uuid_from_url(request, "samples")



"""Tests upload command of Gencove CLI."""
# pylint: disable=wrong-import-order, import-error

from uuid import uuid4

import pytest
from click.testing import CliRunner  # noqa: I100,I201
from gencove.client import APIClient, APIClientError  # noqa: I201
from gencove.command.s3_imports.autoimports.create.cli import create
from gencove.models import S3AutoimportTopic
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/s3_imports/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
        ],
        "before_record_response": [
            filter_jwt,
        ],
    }


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_s3_autoimport_create_project_id_not_uuid(
    credentials,
    mocker,
):
    """Test that project id is valid UUID when creating an automated import."""
    runner = CliRunner()

    mocked_autoimport_from_s3 = mocker.patch.object(
        APIClient,
        "autoimport_from_s3",
    )
    res = runner.invoke(
        create,
        [
            "1234",
            "identifier",
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_autoimport_from_s3.assert_not_called()
    assert "ERROR: Project ID is not valid. Exiting." in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_s3_autoimport_create_invalid_metadata(
    credentials, mocker, project_id
):
    """Test that passed optional metadata is valid when creating an
    automated import.
    """
    runner = CliRunner()

    mocked_autoimport_from_s3 = mocker.patch.object(
        APIClient,
        "autoimport_from_s3",
    )
    res = runner.invoke(
        create,
        [
            project_id,
            "identifier",
            "--metadata-json",
            "[1,2,3",
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_autoimport_from_s3.assert_not_called()
    assert "ERROR: Metadata JSON is not valid. Exiting." in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_s3_autoimport_create_no_permission(credentials, mocker, project_id):
    """Test that user cannot create an automated import if no permissions."""
    runner = CliRunner()

    mocked_autoimport_from_s3 = mocker.patch.object(
        APIClient,
        "autoimport_from_s3",
        side_effect=APIClientError(message="", status_code=403),
    )
    res = runner.invoke(
        create,
        [
            project_id,
            "identifier",
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_autoimport_from_s3.assert_called_once()
    assert "There was an error creating an import job" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_s3_autoimport_create_with_empty_metadata(
    credentials, mocker, project_id
):
    """Test that user can pass empty metadata when creating an
    automated import.
    """
    runner = CliRunner()

    autoimport_from_s3 = S3AutoimportTopic(
        id=uuid4(), topic_arn="arn:aws:sns:us-east-1:123456789012:MyTopic"
    )

    mocked_autoimport_from_s3 = mocker.patch.object(
        APIClient,
        "autoimport_from_s3",
        return_value=autoimport_from_s3,
    )
    res = runner.invoke(
        create,
        [
            project_id,
            "s3://bucket/path/to/project/",
            "--metadata-json",
            None,
            *credentials,
        ],
    )

    assert res.exit_code == 0
    assert (
        "\t".join(
            [
                str(autoimport_from_s3.id),
                autoimport_from_s3.topic_arn,
            ]
        )
        in res.output
    )
    mocked_autoimport_from_s3.assert_called_once()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_s3_autoimport_create_with_metadata(credentials, mocker, project_id):
    """Test that user can pass optional metadata when creating an
    automated import.
    """
    runner = CliRunner()

    autoimport_from_s3 = S3AutoimportTopic(
        id=uuid4(), topic_arn="arn:aws:sns:us-east-1:123456789012:MyTopic"
    )

    mocked_autoimport_from_s3 = mocker.patch.object(
        APIClient,
        "autoimport_from_s3",
        return_value=autoimport_from_s3,
    )
    res = runner.invoke(
        create,
        [
            project_id,
            "s3://bucket/path/to/project/",
            "--metadata-json",
            "[1,2,3]",
            *credentials,
        ],
    )

    assert res.exit_code == 0
    assert (
        "\t".join(
            [
                str(autoimport_from_s3.id),
                autoimport_from_s3.topic_arn,
            ]
        )
        in res.output
    )
    mocked_autoimport_from_s3.assert_called_once()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_s3_autoimport_create_without_metadata(
    credentials, mocker, project_id
):
    """Test that user can create an import job without passing metadata."""
    runner = CliRunner()

    autoimport_from_s3 = S3AutoimportTopic(
        id=uuid4(), topic_arn="arn:aws:sns:us-east-1:123456789012:MyTopic"
    )

    mocked_autoimport_from_s3 = mocker.patch.object(
        APIClient,
        "autoimport_from_s3",
        return_value=autoimport_from_s3,
    )
    res = runner.invoke(
        create,
        [
            project_id,
            "s3://bucket/path/to/project/",
            *credentials,
        ],
    )

    assert res.exit_code == 0
    assert (
        "\t".join(
            [
                str(autoimport_from_s3.id),
                autoimport_from_s3.topic_arn,
            ]
        )
        in res.output
    )
    mocked_autoimport_from_s3.assert_called_once()



"""Project import Biosamples from BaseSpace projects shell command definition.
"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials
from gencove.logger import echo_debug

from .constants import BaseSpaceImportOptionals
from .main import BaseSpaceImport


@click.command("import")
@click.argument("basespace_project_ids")
@click.argument("project_id")
@click.option(
    "--metadata-json",
    required=False,
    default=None,
    help=(
        "Add metadata to all samples that are to be imported from "
        "BaseSpace to a project."
    ),
)
@add_options(common_options)
def basespace_import(  # pylint: disable=too-many-arguments
    basespace_project_ids,
    project_id,
    metadata_json,
    host,
    email,
    password,
    api_key,
):  # pylint: disable=line-too-long
    """Import all Biosamples from BaseSpace projects to a project. Optionally add
    metadata to the samples.

    Examples:

        Import Biosamples to a project:

            gencove basespace projects import 12345678 06a5d04b-526a-4471-83ba-fb54e0941758

        Import Biosamples from multiple BaseSpace projects to a project:

            gencove basespace projects import 12345678,87654321 06a5d04b-526a-4471-83ba-fb54e0941758

        Import Biosamples to a project with metadata:

            gencove basespace projects import 12345678 06a5d04b-526a-4471-83ba-fb54e0941758 --metadata-json='{"batch": "batch1"}'
    """  # noqa: E501
    basespace_project_ids = [
        basespace_project_id.strip()
        for basespace_project_id in basespace_project_ids.split(",")
    ]
    echo_debug(
        "BaseSpace project ids translation: {}".format(basespace_project_ids)
    )

    BaseSpaceImport(
        basespace_project_ids,
        project_id,
        Credentials(email=email, password=password, api_key=api_key),
        BaseSpaceImportOptionals(host=host, metadata_json=metadata_json),
    ).run()



"""Utilities for processing batch types."""


def get_line(batch_type):
    """Build a list of relevant data to be printed.

    Args:
        batch_type (dict): an object from project batch types response

    Returns:
        str: tab delimited list of relevant data to be printed
    """
    return "\t".join(
        [
            batch_type.key,
            batch_type.description,
        ]
    )



"""Util functions."""
import json
import operator

MOCK_UUID = "11111111-1111-1111-1111-111111111111"


def get_vcr_response(url, vcr, matches=operator.eq, just_body=True):
    """Returns the VCR response of the first request that matches the url.

    Args:
        url (str): The request URL you want to match
        vcr (Cassette): VCR object, use the fixture
        matches (callable): Evaluator function
        just_body (bool): Set to False to retrieve the entire response object.

    Returns:
        response (dict): The VCR response.
    """
    request = next(
        request for request in vcr.requests if matches(request.path, url)
    )
    response = vcr.responses_of(request)[0]
    if just_body:
        try:
            response = json.loads(response["body"]["string"])
        except json.decoder.JSONDecodeError:
            pass
    return response



"""Project batch types list shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals
from gencove.logger import echo_debug

from .main import CreateBatch


@click.command("create-batch")
@click.argument("project_id")
@click.option(
    "--batch-type",
    help="One of available project's batch types.\n"
    "Use `gencove projects list-batch-types` command to find out "
    "which batch types are available.",
)
@click.option("--batch-name", help="User defined batch name.")
@click.option(
    "--sample-ids",
    default="",
    help="A comma separated list of sample ids for "
    "which to create a batch; if not specified use all samples in project",
)
@add_options(common_options)
def create_project_batch(  # pylint: disable=too-many-arguments
    project_id,
    batch_type,
    batch_name,
    sample_ids,
    host,
    email,
    password,
    api_key,
):
    """Create a batch in a project."""
    if sample_ids:
        sample_ids = [s_id.strip() for s_id in sample_ids.split(",")]
        echo_debug("Sample ids translation: {}".format(sample_ids))
    else:
        sample_ids = []

    CreateBatch(
        project_id,
        batch_type,
        batch_name,
        sample_ids,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Constants for basespace-autoimport command."""
from typing import Optional

from gencove.constants import Optionals


class BaseSpaceAutoImportOptionals(Optionals):
    """BaseSpaceAutoImportOptionals model"""

    metadata_json: Optional[str]



"""Test download command."""
# pylint: disable=wrong-import-order, import-error
import operator
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import APIClient
from gencove.command.samples.cli import download_file
from gencove.models import SampleDetails
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import (
    filter_aws_headers,
    filter_jwt,
    filter_samples_request,
    filter_samples_response,
    mock_binary_response,
    replace_gencove_url_vcr,
    replace_s3_from_url,
)
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/samples/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            replace_s3_from_url,
            filter_samples_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_aws_headers,
            mock_binary_response,
            filter_samples_response,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_samples_download_file_local(
    credentials, mocker, recording, sample_id_download, vcr
):
    """Test command outputs to local destination."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        file_type = "fastq-r1"
        file_path = "r1.fastq.gz"
        if not recording:
            # Mock only if using the cassettes, since we mock the return value.
            get_sample_details_response = get_vcr_response(
                "/api/v2/samples/", vcr, operator.contains
            )
            mocked_sample_details = mocker.patch.object(
                APIClient,
                "get_sample_details",
                return_value=SampleDetails(**get_sample_details_response),
            )
            file_content = get_vcr_response(f"/{file_path}", vcr)
            file_content = file_content["body"]["string"]
        res = runner.invoke(
            download_file,
            [
                sample_id_download,
                file_type,
                file_path,
                *credentials,
            ],
        )
        assert res.exit_code == 0
        if not recording:
            mocked_sample_details.assert_called_once()
            with open(file_path, "rb") as local_file:
                assert file_content == local_file.read()


@pytest.mark.vcr
@assert_authorization
def test_samples_download_file_stdout(
    credentials, mocker, recording, sample_id_download, vcr
):
    """Test command outputs to stdout."""
    runner = CliRunner()
    file_type = "fastq-r1"
    if not recording:
        # Mock only if using the cassettes, since we mock the return value.
        get_sample_details_response = get_vcr_response(
            "/api/v2/samples/", vcr, operator.contains
        )
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value=SampleDetails(**get_sample_details_response),
        )
        file_content = get_vcr_response("/r1.fastq.gz", vcr)
        file_content = file_content["body"]["string"]

    res = runner.invoke(
        download_file,
        [
            sample_id_download,
            file_type,
            "-",
            *credentials,
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_sample_details.assert_called_once()
        assert file_content == res.output.encode()


@pytest.mark.vcr
def test_samples_download_file_directory(mocker):
    """Test command fails when writing to directory."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(APIClient, "login")
        res = runner.invoke(
            download_file,
            [
                str(uuid4()),
                "file_type",
                ".",
                "--email",
                "foo@bar.com",
                "--password",
                "12345",
            ],
        )
        assert res.exit_code == 1
        mocked_login.assert_not_called()
        assert (
            "ERROR: Please specify a file path (not directory path) for DESTINATION"  # noqa: E501 line too long pylint: disable=line-too-long
            in res.output
        )



"""Tests for utils of Gencove CLI."""
import csv
import os
from enum import Enum

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError
from gencove.command.download.utils import (
    _get_prefix_parts,
    build_file_path,
    get_download_template_format_params,
)
from gencove.command.upload.utils import (
    _validate_header,
    parse_fastqs_map_file,
    upload_file,
)
from gencove.command.utils import is_valid_uuid
from gencove.constants import (
    ApiEndpoints,
    Credentials,
    DOWNLOAD_TEMPLATE,
    DownloadTemplateParts,
)
from gencove.exceptions import ValidationError
from gencove.utils import enum_as_dict, login


def test_upload_file(mocker):
    """Sanity check upload function."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("foo.txt", "w") as fastq_file:
            fastq_file.write("AAABBB")
        mocked_s3_client = mocker.Mock()
        assert upload_file(
            mocked_s3_client,
            "foo.txt",
            "foo-bucket",
            object_name="foo-object.txt",
        )
        mocked_s3_client.upload_file.assert_called_once()


def test_download_template_tokens():
    """Ensure download tokens are only the ones defined."""
    assert [
        "client_id",
        "gencove_id",
        "file_type",
        "file_extension",
        "default_filename",
    ] == list(enum_as_dict(DownloadTemplateParts).values())


# pylint: disable=too-many-locals
def test_build_file_path():
    """Test token combinations when building a file path."""
    client_id = "12345"
    gencove_id = "1"
    deliverable = {
        "download_url": "https://example.com/file.txt",
        "file_type": "txt",
    }
    # file with prefix will never have client_id and gencove_id. 0 is default
    file_with_prefix0 = "{client_id}/{gencove_id}/{default_filename}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    file_with_prefix1 = "{client_id}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    file_with_prefix2 = "{gencove_id}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    file_with_prefix3 = "{file_type}"
    file_with_prefix4 = "{file_extension}"
    file_with_prefix5 = "{default_filename}"
    file_with_prefix6 = "{default_filename}.{file_extension}"
    file_with_prefix7 = "{default_filename}.{client_id}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    file_with_prefix8 = "{client_id}.{default_filename}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    file_with_prefix9 = "{file_extension}_{client_id}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    download_to = "."
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = build_file_path(deliverable, file_with_prefix0, download_to)
        # default_filename is always a full filename with any extensions
        assert result == "./{}/{}/{}".format(
            client_id, gencove_id, "file.txt"
        )
        result = build_file_path(deliverable, file_with_prefix1, download_to)
        assert result == "./{}".format(client_id)
        result = build_file_path(deliverable, file_with_prefix2, download_to)
        assert result == "./{}".format(gencove_id)
        result = build_file_path(deliverable, file_with_prefix3, download_to)
        assert result == "./{}".format("txt")
        result = build_file_path(deliverable, file_with_prefix4, download_to)
        assert result == "./{}".format("txt")
        result = build_file_path(deliverable, file_with_prefix5, download_to)
        assert result == "./{}".format("file.txt")
        result = build_file_path(deliverable, file_with_prefix6, download_to)
        assert result == "./{}.{}".format("file.txt", "txt")
        result = build_file_path(deliverable, file_with_prefix7, download_to)
        assert result == "./{}.{}".format("file.txt", client_id)
        result = build_file_path(deliverable, file_with_prefix8, download_to)
        assert result == "./{}.{}".format(client_id, "file.txt")
        result = build_file_path(deliverable, file_with_prefix9, download_to)
        assert result == "./{}_{}".format("txt", client_id)


def test___get_filename_dirs_prefix():
    """Test proper processing of parts in download template."""
    client_id = "12345"
    gencove_id = "1"
    template = DOWNLOAD_TEMPLATE.format(
        **get_download_template_format_params(client_id, gencove_id)
    )

    resp = _get_prefix_parts(template)

    assert resp.dirs == "{}/{}".format(client_id, gencove_id)
    assert resp.filename == "{{{}}}".format(
        DownloadTemplateParts.DEFAULT_FILENAME.value
    )
    assert resp.file_extension == ""

    template2 = "{client_id}-{gencove_id}_{file_type}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    resp = _get_prefix_parts(template2)
    assert resp.dirs == ""
    assert resp.filename == "{}-{}_{{{}}}".format(
        client_id, gencove_id, DownloadTemplateParts.FILE_TYPE.value
    )
    assert resp.file_extension == ""

    template3 = "{client_id}-{gencove_id}_{file_type}.vcf.gz".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    resp = _get_prefix_parts(template3)
    assert resp.dirs == ""
    assert resp.filename == "{}-{}_{{{}}}".format(
        client_id, gencove_id, DownloadTemplateParts.FILE_TYPE.value
    )
    assert resp.file_extension == "vcf.gz"


def test_parse_fastqs_map_file():
    """Test parsing of map file into dict."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/test.fastq.gz", "w") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/test_2.fastq.gz", "w") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_dir/test_3.fastq.gz", "w") as fastq_file3:
            fastq_file3.write("AAABBB")

        with open("test_map.csv", "w") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r2", "test_dir/test_2.fastq.gz"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                ]
            )

        fastqs = parse_fastqs_map_file("test_map.csv")
        assert len(fastqs) == 2
        assert ("barid", "R1") in fastqs
        assert ("barid", "R2") in fastqs
        assert len(fastqs[("barid", "R1")]) == 2
        assert fastqs[("barid", "R1")][0] == "test_dir/test.fastq.gz"
        assert fastqs[("barid", "R1")][1] == "test_dir/test_3.fastq.gz"


def test_invalid_fastqs_map_file():
    """Test that error is raised if file is invalid."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/test.fastq.gz", "w") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/test_2.fastq.gz", "w") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_dir/test_3.fastq.gz", "w") as fastq_file3:
            fastq_file3.write("AAABBB")

        with open("test_map.csv", "w") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r2", "test_dir/test_2.fastq.zip"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                ]
            )
        try:
            parse_fastqs_map_file("test_map.csv")
        except ValidationError as err:
            assert "Bad file extension in path" in err.args[0]


def test_fastqs_map_file_path_does_not_exist():
    """Test that error is raised if file is invalid."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/test.fastq.gz", "w") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/test_2.fastq.gz", "w") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_dir/test_3.fastq.gz", "w") as fastq_file3:
            fastq_file3.write("AAABBB")

        with open("test_map.csv", "w") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                    ["barid", "r1", "test_dir/i_dont_exist.fastq.gz"],
                ]
            )
        try:
            parse_fastqs_map_file("test_map.csv")
        except ValidationError as err:
            assert "Could not find" in err.args[0]


def test_fastqs_map_file_has_wrong_header():
    """Test that header is validated properly."""
    runner = CliRunner()
    # This context manager is used to avoid unnecessarily creating a
    # test_map.csv file in the root of the project on every test run
    with runner.isolated_filesystem():
        with open("test_map.csv", "w") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r1_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                    ["barid", "r1", "test_dir/i_dont_exist.fastq.gz"],
                ]
            )
        try:
            parse_fastqs_map_file("test_map.csv")
        except ValidationError as err:
            assert "Unexpected CSV header" in err.args[0]


def test__validate_header():
    """Test that header is validated properly."""
    header_row = dict(foo="foo", bar="bar")
    try:
        _validate_header(header_row)
    except ValidationError:
        pass

    header_row = dict(
        client_id="client_id", r_notation="r_notation", path="path"
    )
    assert _validate_header(header_row) is None


def test_is_valid_uuid__is_valid():
    """Test that a UUID is a valid UUID"""
    assert is_valid_uuid("11111111-1111-1111-1111-111111111111")


def test_is_valid_uuid__is_not_valid__too_long():
    """Test that UUID with extra chars is not a valid UUID"""
    assert is_valid_uuid("11111111-1111-1111-1111-11111111111122") is False


def test_is_valid_uuid__is_not_valid__too_short():
    """Test that UUID with missing chars is not a valid UUID"""
    assert is_valid_uuid("11111111-1111-1111-1111-1") is False


def test_is_valid_uuid__is_not_valid__text():
    """Test that random word is not a valid UUID"""
    assert is_valid_uuid("foo") is False


def test_enum_as_dict():
    """Test that the function enum_as_dict returns properly formatted dict."""

    class TestEnum(Enum):
        """Enum for testing"""

        KEY1 = "key1"
        KEY2 = "key2"
        KEY3 = "key3"

    assert enum_as_dict(TestEnum) == {
        "KEY1": "key1",
        "KEY2": "key2",
        "KEY3": "key3",
    }


def test_login_mfa(mocker):
    """Test that the the login function in APIClient asks for the one time
    password token.
    """
    api_client = APIClient()
    credentials = Credentials(
        email="foo@bar.com", password="123456", api_key=""
    )

    def _request(
        endpoint,
        params,
        *args,  # pylint: disable=unused-argument
        **kwargs,  # pylint: disable=unused-argument
    ):
        if endpoint == ApiEndpoints.GET_JWT.value:
            if "otp_token" not in params:
                raise APIClientError(
                    {"otp_token": ["Please enter your OTP token."]}, 401
                )
            return {"access": "access", "refresh": "refresh"}
        return {}

    mocked_request = mocker.patch.object(
        APIClient, "_request", side_effect=_request
    )
    mocked_prompt = mocker.patch(
        "gencove.utils.click.prompt", return_value="token"
    )

    login(api_client, credentials)
    assert mocked_request.call_count == 2
    mocked_prompt.assert_called_once_with(
        "One time password", type=str, err=True
    )



"""List Biosamples from BaseSpace project shell command definition.
"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import BiosamplesList


@click.command("list")
@click.argument("basespace_project_id")
@add_options(common_options)
def biosamples_list(
    basespace_project_id,
    host,
    email,
    password,
    api_key,
):
    """List all Biosamples from BaseSpace project.

    Examples:

        List Biosamples of a BaseSpace project:

            gencove basespace biosamples list 12345678

        List Biosamples of a BaseSpace projects:

            gencove basespace biosamples list 12345678,87654321

    """
    BiosamplesList(
        basespace_project_id,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""List projects subcommand."""
import backoff

# pylint: disable=wrong-import-order
from gencove.client import APIClientError, APIClientTimeout  # noqa: I100
from gencove.command.base import Command

from .utils import get_line
from ...utils import is_valid_uuid
from ....exceptions import ValidationError


class ListSamples(Command):
    """List samples command executor."""

    def __init__(self, project_id, credentials, options):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.sample_status = options.status
        self.sample_archive_status = options.archive_status
        self.search_term = options.search

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""
        if is_valid_uuid(self.project_id) is False:
            raise ValidationError("Project ID is not valid. Exiting.")

    def execute(self):
        self.echo_debug(
            "Retrieving sample sheet: "
            "status={} archive_status={} search_term={}".format(
                self.sample_status,
                self.sample_archive_status,
                self.search_term,
            )
        )
        try:
            for samples in self.get_paginated_samples():
                if not samples:
                    self.echo_debug("No matching samples were found.")
                    return

                for sample in samples:
                    self.echo_data(get_line(sample))
        except APIClientError as err:
            if err.status_code == 404:
                self.echo_error(
                    "Project {} does not exist.".format(self.project_id)
                )
            raise

    def get_paginated_samples(self):
        """Paginate over all sample sheets for the destination.

        Yields:
            paginated lists of samples
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get sample sheet page")
            resp = self.get_samples(next_link)
            yield resp.results
            next_link = resp.meta.next
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_samples(self, next_link=None):
        """Get sample sheet page."""
        return self.api_client.get_project_samples(
            project_id=self.project_id,
            next_link=next_link,
            search=self.search_term,
            sample_status=self.sample_status,
            sample_archive_status=self.sample_archive_status,
        )



"""Test project list samples command."""
# pylint: disable=wrong-import-order, import-error
import io
import operator
import sys
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError, APIClientTimeout
from gencove.command.projects.cli import list_project_samples
from gencove.logger import echo_data
from gencove.models import ProjectSamples, SampleDetails
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_get_project_samples_request,
    filter_get_project_samples_response,
)
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_get_project_samples_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_get_project_samples_response,
        ],
    }


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_empty(mocker, credentials, project_id):
    """Test project has no samples."""
    runner = CliRunner()
    mocked_get_project_samples = mocker.patch.object(
        APIClient,
        "get_project_samples",
        return_value=ProjectSamples(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(
        list_project_samples,
        [project_id, *credentials],
    )
    assert res.exit_code == 0
    mocked_get_project_samples.assert_called_once()
    assert res.output == ""


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_projects_bad_project_id(
    mocker, credentials
):  # pylint: disable=unused-argument
    """Test project samples throw an error if project id is not uuid."""
    runner = CliRunner()

    project_id = "123456"

    res = runner.invoke(
        list_project_samples,
        [project_id, *credentials],
    )
    assert res.exit_code == 1
    output_line = io.BytesIO()
    sys.stdout = output_line
    echo_data(
        "\n".join(
            [
                "ERROR: Project ID is not valid. Exiting.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_projects_no_project(mocker, credentials):
    """Test project samples throw an error if no project available."""
    runner = CliRunner()
    mocked_get_project_samples = mocker.patch.object(
        APIClient,
        "get_project_samples",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=404
        ),
        return_value={"detail": "Not found"},
    )

    project_id = str(uuid4())

    res = runner.invoke(
        list_project_samples,
        [project_id, *credentials],
    )
    assert res.exit_code == 1
    mocked_get_project_samples.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo_data(
        "\n".join(
            [
                "ERROR: Project {} does not exist.".format(project_id),
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_project_samples_slow_response_retry(mocker, credentials):
    """Test project samples slow response retry."""
    runner = CliRunner()
    mocked_get_project_samples = mocker.patch.object(
        APIClient,
        "get_project_samples",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )

    res = runner.invoke(
        list_project_samples,
        [str(uuid4()), *credentials],
    )
    assert res.exit_code == 1
    assert mocked_get_project_samples.call_count == 2


@pytest.mark.vcr
@assert_authorization
def test_list_project_samples(
    mocker, credentials, project_id, recording, vcr
):  # pylint: disable=unused-argument
    """Test project samples being outputed to the shell."""
    runner = CliRunner()

    if not recording:
        # Mock get_project_samples only if using the cassettes,
        # since we mock the return value.
        mocked_samples = get_vcr_response(
            "/api/v2/project-samples/", vcr, operator.contains
        )
        mocked_get_project_samples = mocker.patch.object(
            APIClient,
            "get_project_samples",
            return_value=ProjectSamples(**mocked_samples),
        )
    res = runner.invoke(
        list_project_samples,
        [project_id, *credentials],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_get_project_samples.assert_called_once()
        output_line = io.BytesIO()
        sys.stdout = output_line
        for mocked_sample in mocked_samples["results"]:
            mocked_sample = SampleDetails(**mocked_sample)
            echo_data(
                "\t".join(
                    [
                        str(mocked_sample.last_status.created.isoformat()),
                        str(mocked_sample.id),
                        mocked_sample.client_id,
                        mocked_sample.last_status.status,
                        mocked_sample.archive_last_status.status,
                    ]
                )
            )
        assert output_line.getvalue() == res.output.encode()



"""Utilities for processing projects."""
from gencove.command.utils import sanitize_string


def get_line(project):
    """Build a list of relevant data to be printed.

    Args:
        project (Project): instance of project

    Returns:
        list(str): list of relevant data to be printed
    """
    return "\t".join(
        [
            str(project.created),
            str(project.id),
            sanitize_string(project.name),
            sanitize_string(project.pipeline_capabilities.name),
        ]
    )



"""Assign uploads from a prefix to a project subcommand."""

import json

import backoff

from ...base import Command
from ...utils import is_valid_json, is_valid_uuid
from .... import client
from ....constants import ASSIGN_BATCH_SIZE, UPLOAD_PREFIX
from ....exceptions import ValidationError
from ....utils import batchify


class RunPrefix(Command):
    """Run prefix command executor."""

    def __init__(self, project_id, prefix, credentials, options):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.prefix = prefix
        self.metadata_json = options.metadata_json
        self.status = options.status

    def initialize(self):
        """Initialize run-prefix subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if is_valid_uuid(self.project_id) is False:
            raise ValidationError("Project ID is not valid. Exiting.")
        if self._valid_prefix() is False:
            raise ValidationError("Prefix is not valid. Exiting.")
        if self.metadata_json and is_valid_json(self.metadata_json) is False:
            raise ValidationError("Metadata JSON is not valid. Exiting.")

    def execute(self):
        """Assign samples to the project from the prefixed path."""
        try:
            self.echo_debug(
                "Retrieving sample sheet: search_term={}".format(self.prefix)
            )
            samples = []
            self.echo_info("Gathering uploads.")
            for sample_sheet in self._get_paginated_sample_sheet():
                if not sample_sheet:
                    raise ValidationError("No matching paths found.")
                samples.extend(sample_sheet)
            self._assign_samples(samples)
            self.echo_info(
                "Number of samples assigned to the project {}: {}".format(
                    self.project_id, len(samples)
                )
            )
        except client.APIClientError:  # pylint: disable=try-except-raise
            raise

    def _valid_prefix(self):
        """Validate if passed prefix starts with the correct sequence."""
        if not self.prefix.startswith(UPLOAD_PREFIX):
            self.echo_error(
                "Invalid destination path. Must start with {}".format(
                    UPLOAD_PREFIX
                )
            )
            return False
        return True

    # duplicated in upload/main.py with the exception difference
    def _get_paginated_sample_sheet(self):
        """Paginate over all sample sheet pages.

        Yields:
            paginated lists of uploads
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get sample sheet page")
            resp = self._get_sample_sheet(next_link)
            yield resp.results
            next_link = resp.meta.next
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=5,
        max_time=30,
    )
    def _get_sample_sheet(self, next_link=None):
        """Get sample sheet page."""
        return self.api_client.get_sample_sheet(
            gncv_path=self.prefix,
            assigned_status=self.status,
            next_link=next_link,
        )

    def _assign_samples(self, samples):
        """Assign samples to the project optionally passing metadata."""
        try:
            # prepare metadata
            metadata = None
            if self.metadata_json is not None:
                metadata = json.loads(self.metadata_json)
                self.echo_info("Assigning metadata to the uploaded samples.")
            assigned_count = 0
            for samples_batch in batchify(
                samples, batch_size=ASSIGN_BATCH_SIZE
            ):
                try:
                    samples_batch_len = len(samples_batch)
                    self.echo_info(
                        "Assigning batch: {}".format(samples_batch_len)
                    )
                    self.api_client.add_samples_to_project(
                        samples_batch, self.project_id, metadata
                    )
                    assigned_count += samples_batch_len
                    self.echo_info(
                        "Total assigned: {}".format(assigned_count)
                    )
                except client.APIClientError as err:
                    self.echo_debug(err)
                    self.echo_error(
                        "There was an error assigning/running samples."
                    )
                    if assigned_count > 0:
                        self.echo_warning(
                            "Some of the samples were assigned. "
                            "Please use the Web UI to assign "
                            "the rest of the samples."
                        )
                    else:
                        self.echo_error(
                            "There was an error assigning samples."
                        )
                    raise
            self.echo_info("Assigned all samples to a project.")

        except client.APIClientError as err:
            self.echo_debug(err.message)
            self.echo_error("There was an error assigning samples.")
            raise



"""List BaseSpace AutoImport jobs subcommand."""
import backoff

from .utils import get_line
from ....base import Command
from ..... import client


class BaseSpaceAutoImportList(Command):
    """BaseSpace autoimport list command executor."""

    def initialize(self):
        """Initialize basespace-autoimport subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    def execute(self):
        """Make a request to list jobs to periodically import Biosamples from
        BaseSpace projects.
        """
        self.echo_debug("List AutoImport jobs.")

        try:
            for (
                basespace_autoimports
            ) in self.get_paginated_basespace_autoimports():
                if not basespace_autoimports:
                    self.echo_debug(
                        "No BaseSpace autoimport jobs were found."
                    )
                    return
                for basespace_autoimport in basespace_autoimports:
                    self.echo_data(get_line(basespace_autoimport))
        except client.APIClientError:
            self.echo_error(
                "There was an error listing autoimport jobs of BaseSpace "
                "projects."
            )
            raise

    def get_paginated_basespace_autoimports(self):
        """Paginate over all BaseSpace autoimports.

        Yields:
            paginated lists of BaseSpace autoimports
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get BaseSpace autoimport page")
            resp = self.get_basespace_autoimports(next_link)

            yield resp.results
            next_link = resp.meta.next
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_basespace_autoimports(self, next_link=None):
        """Get BaseSpace autoimports page."""
        return self.api_client.list_basespace_autoimport_jobs(next_link)



"""Download command executor."""
import json
import re

import backoff

import requests

from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.command.download.exceptions import DownloadTemplateError
from gencove.constants import SampleArchiveStatus
from gencove.exceptions import ValidationError

from .constants import (
    ALLOWED_ARCHIVE_STATUSES_RE,
    ALLOWED_STATUSES_RE,
    METADATA_FILE_TYPE,
    QC_FILE_TYPE,
)
from .utils import (
    build_file_path,
    download_file,
    fatal_process_sample_error,
    get_download_template_format_params,
    save_metadata_file,
    save_qc_file,
)


# pylint: disable=too-many-instance-attributes
class Download(Command):
    """Download command executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        download_to,
        filters,
        credentials,
        options,
        download_urls,
        no_progress,
    ):
        super().__init__(credentials, options)
        self.download_to = download_to
        self.filters = filters
        self.options = options
        self.sample_ids = set()
        self.downloaded_files = set()
        self.download_urls = download_urls
        self.download_files = []
        self.no_progress = no_progress

    def initialize(self):
        """Initialize download command."""
        if self.filters.project_id and self.filters.sample_ids:
            raise ValidationError("Bad configuration. Exiting")

        self.login()

        if self.filters.project_id:
            self.echo_debug(
                "Retrieving sample ids for a project: {}".format(
                    self.filters.project_id
                )
            )

            try:
                samples_generator = self._get_paginated_samples()
                for sample in samples_generator:
                    self.sample_ids.add(sample.id)
            except client.APIClientError as err:
                raise ValidationError(
                    "Project id {} not found.".format(self.filters.project_id)
                ) from err
        else:
            self.sample_ids = self.filters.sample_ids

    def validate(self):
        """Validate command configuration before execution.

        Raises:
            ValidationError : something is wrong with configuration
        """
        if not self.filters.project_id and not self.filters.sample_ids:
            raise ValidationError(
                "Must specify one of: project id or sample ids"
            )

        if self.filters.project_id and self.filters.sample_ids:
            raise ValidationError(
                "Must specify only one of: project id or sample ids"
            )

        if not self.sample_ids:
            raise ValidationError("No samples to process. Exiting.")

        if all(
            [
                self.download_to == "-",
                not self.download_urls,
            ]
        ):
            raise ValidationError(
                "Cannot have - as a destination without download-urls."
            )

        if self.download_urls:
            self.echo_debug("Requesting download list in JSON format.")
        if self.download_to == "-":
            self.echo_debug("Will not download. Redirecting to STDOUT.")
        else:
            self.echo_debug(
                "Host is {} saving to {}".format(
                    self.options.host, self.download_to
                )
            )

    def execute(self):
        if self.download_to != "-":
            self.echo_info("Processing samples")
        for sample_id in self.sample_ids:
            try:
                self.process_sample(sample_id)
            except DownloadTemplateError:
                return
        if self.download_urls:
            self.output_list()

    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.HTTPError,
        giveup=fatal_process_sample_error,
        max_tries=10,
    )
    def process_sample(self, sample_id):
        """Process sample.

        Check if a sample is in appropriate state and if it is,
        get its files one by one.

        If downloading and a download failed with error 403, reprocess the
        sample in order to get fresh download url.
        """
        try:
            sample = self.api_client.get_sample_details(sample_id)
        except client.APIClientError:
            self.echo_warning(
                "Sample with id {} not found. "
                "Are you using client id instead of sample id?".format(
                    sample_id
                )
            )
            return

        self.echo_debug(
            "Processing sample id {}, status {}".format(
                sample.id, sample.last_status.status
            )
        )

        if not ALLOWED_ARCHIVE_STATUSES_RE.match(
            sample.archive_last_status.status
        ):
            raise ValidationError(
                "Sample with id {} is archived and cannot be downloaded - "
                "please restore the sample and try again.".format(sample.id)
            )

        if not ALLOWED_STATUSES_RE.match(sample.last_status.status):
            self.echo_warning(
                "Sample #{} has no deliverable.".format(sample.id),
            )
            return

        file_types_re = re.compile(
            "|".join(self.filters.file_types), re.IGNORECASE
        )

        file_with_prefix = self.options.download_template.format(
            **get_download_template_format_params(sample.client_id, sample.id)
        )
        self.echo_debug(
            "file path with prefix is: {}".format(file_with_prefix)
        )
        if all(
            [
                self.download_to != "-",
                not self.download_urls,
                file_types_re.match(QC_FILE_TYPE),
            ]
        ):
            self.download_sample_qc_metrics(file_with_prefix, sample_id)

        if all(
            [
                self.download_to != "-",
                not self.download_urls,
                file_types_re.match(METADATA_FILE_TYPE),
            ]
        ):
            self.download_sample_metadata(file_with_prefix, sample_id)

        self.download_files.append(
            {
                "gencove_id": sample.id,
                "client_id": sample.client_id,
                "last_status": {
                    "id": sample.last_status.id,
                    "status": sample.last_status.status,
                    "created": sample.last_status.created,
                },
                "archive_last_status": {
                    "id": sample.archive_last_status.id,
                    "status": sample.archive_last_status.status,
                    "created": sample.archive_last_status.created,
                    "transition_cutoff": (
                        sample.archive_last_status.transition_cutoff
                    ),
                },
                "files": {},
            }
        )

        for sample_file in sample.files:
            # pylint: disable=E0012,C0330
            if self.filters.file_types and not file_types_re.match(
                sample_file.file_type
            ):
                self.echo_debug(
                    "Deliverable file type is not in desired file types"
                )
                continue

            if not self.download_urls:
                file_path = build_file_path(
                    sample_file, file_with_prefix, self.download_to
                )
                self.validate_and_download(
                    file_path,
                    download_file,
                    file_path,
                    sample_file.download_url,
                    self.options.skip_existing,
                    self.no_progress,
                )
            self.download_files[-1]["files"][sample_file.file_type] = {
                "id": sample_file.id,
                "download_url": sample_file.download_url,
            }

    def validate_and_download(
        self, download_to_path, download_func, *args, **kwargs
    ):
        """Check if this file was already downloaded, if yes - exit.

        Args:
            download_to_path(str): system file path to donwload to
            download_func(function): function that will do the download logic
            *args: arguments that will be passed to download function
            **kwargs: keyword arguments that will be passed to download func

        Returns:
            None

        Raises:
            DownloadTemplateError: if the file was found in already downloaded
             list
        """
        if download_to_path in self.downloaded_files:
            self.echo_warning(
                "Bad template! Multiple files have the same name. "
                "Please fix the template and try again."
            )

            raise DownloadTemplateError

        download_func(*args, **kwargs)

        self.echo_debug("Adding file path: {}".format(download_to_path))
        self.downloaded_files.add(download_to_path)

    def download_sample_qc_metrics(self, file_with_prefix, sample_id):
        """Download and save to file on user file system.

        Args:
            file_with_prefix(str): file path based on download template,
                prefilled with sample data
            sample_id(str of uuid): sample gencove id

        Returns:
            None
        """
        file_path = build_file_path(
            dict(file_type=QC_FILE_TYPE),
            file_with_prefix,
            self.download_to,
            "{}_{}.json".format(sample_id, QC_FILE_TYPE),
        )

        self.validate_and_download(
            file_path,
            save_qc_file,
            file_path,
            self.api_client,
            sample_id,
            self.options.skip_existing,
        )

    def download_sample_metadata(self, file_with_prefix, sample_id):
        """Get metadata and save to file on user file system.

        Args:
            file_with_prefix(str): file path based on download template,
                prefilled with sample data
            sample_id(str of uuid): sample gencove id

        Returns:
            None
        """
        file_path = build_file_path(
            dict(file_type=METADATA_FILE_TYPE),
            file_with_prefix,
            self.download_to,
            "{}_{}.json".format(sample_id, METADATA_FILE_TYPE),
        )

        self.validate_and_download(
            file_path,
            save_metadata_file,
            file_path,
            self.api_client,
            sample_id,
            self.options.skip_existing,
        )

    def _get_paginated_samples(self):
        """Generate for project samples that traverses all pages."""
        get_samples = True
        next_page = None
        while get_samples:
            self.echo_debug("Getting page: {}".format(next_page or 1))
            req = self.api_client.get_project_samples(
                self.filters.project_id,
                next_page,
                sample_archive_status=SampleArchiveStatus.AVAILABLE.value,
            )
            for sample in req.results:
                yield sample
            next_page = req.meta.next
            get_samples = next_page is not None

    def output_list(self):
        """Output reformatted JSON of each individual sample."""
        self.echo_debug("Outputting JSON.")
        if self.download_to == "-":
            self.echo_data(
                json.dumps(
                    self.download_files, indent=4, cls=client.CustomEncoder
                )
            )
        else:
            with open(self.download_to, "w") as json_file:
                json_file.write(
                    json.dumps(
                        self.download_files,
                        indent=4,
                        cls=client.CustomEncoder,
                    )
                )
            self.echo_info(
                "Samples and their deliverables download URLs outputted to "
                "{}".format(self.download_to)
            )



"""Uploads commands to be executed from command line."""
import click

from .list.cli import list_uploads


@click.group()
def uploads():
    """Uploads management commands."""


uploads.add_command(list_uploads)



"""Constants for upload command."""
from enum import Enum, unique
from typing import Optional

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from gencove.constants import Optionals  # noqa: I100

# pylint: disable=invalid-name

TMP_UPLOADS_WARNING = (
    "The Gencove upload area is a temporary storage. "
    "Log into the dashboard and assign uploaded files "
    "to a project in order to avoid automatic deletion."
)


@unique
class UploadStatuses(Enum):
    """UploadStatuses enum"""

    DONE = "succeeded"
    STARTED = "started"
    FAILED = "failed"


FASTQ_EXTENSIONS = (".fastq.gz", ".fastq.bgz", ".fq.gz", ".fq.bgz")


# pylint: disable=too-few-public-methods
class UploadOptions(Optionals):
    """UploadOptions model"""

    project_id: Optional[str]
    metadata: Optional[str]


ASSIGN_ERROR = (
    "Your files were successfully uploaded, "
    "but there was an error automatically running them "
    "and assigning them to project id {}."
    "You can try to assign without upload using following gncv path: {}"
)


# pylint: disable=too-few-public-methods
class FastQ(BaseModel):
    """FastQ model"""

    client_id: str
    r_notation: str
    path: str


R_NOTATION_MAP = {"R1": "R1", "R2": "R2", "r1": "R1", "r2": "R2"}


@unique
class PathTemplateParts(Enum):
    """PathTemplateParts enum"""

    client_id = "client_id"
    r_notation = "r_notation"


PATH_TEMPLATE = "{{{}}}_{{{}}}.fastq.gz".format(
    PathTemplateParts.client_id.value, PathTemplateParts.r_notation.value
)



"""Constants for run-prefix command."""
from typing import Optional

from gencove.constants import Optionals, SampleAssignmentStatus


# pylint: disable=too-few-public-methods
class RunPrefixOptionals(Optionals):
    """RunPrefixOptionals model"""

    metadata_json: Optional[str]
    status: Optional[SampleAssignmentStatus]



"""Commands to be executed from command line."""
import click

from .download_file.cli import download_file
from .get_metadata.cli import get_metadata
from .set_metadata.cli import set_metadata


@click.group()
def samples():
    """Sample managements commands."""


samples.add_command(get_metadata)
samples.add_command(set_metadata)
samples.add_command(download_file)



"""Utility to generate version string."""
import pkgutil


def version():
    """Return version string."""
    major = pkgutil.get_data(__package__, "version/A-major").strip()
    minor = pkgutil.get_data(__package__, "version/B-minor").strip()
    patch = pkgutil.get_data(__package__, "version/C-patch").strip()
    return "{}.{}.{}".format(
        major.decode("utf-8"), minor.decode("utf-8"), patch.decode("utf-8")
    )



"""Utilities for processing and validating samples."""


def get_line(sample):
    """Build a list of relevant data to be printed.

    Args:
        sample (dict): an object from project samples

    Returns:
        list(str): list of relevant data to be printed
    """
    return "\t".join(
        [
            sample.last_status.created.isoformat(),
            str(sample.id),
            str(sample.client_id),
            sample.last_status.status,
            sample.archive_last_status.status,
        ]
    )



"""Test project's run prefix command."""

import copy
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
)  # noqa: I100
from gencove.command.projects.cli import run_prefix
from gencove.models import SampleSheet, UploadSamples

MOCKED_UPLOADS = dict(
    meta=dict(next=None),
    results=[
        {
            "client_id": "clientid1",
            "fastq": {
                "r1": {
                    "upload": str(uuid4()),
                    "destination_path": "gncv://batch1/clientid1_R1.fastq.gz",
                    "last_status": {
                        "id": str(uuid4()),
                        "status": "unassigned",
                    },
                },
                "r2": {
                    "upload": str(uuid4()),
                    "destination_path": "gncv://batch1/clientid1_R2.fastq.gz",
                    "last_status": {
                        "id": str(uuid4()),
                        "status": "unassigned",
                    },
                },
            },
        },
        {
            "client_id": "clientid2",
            "fastq": {
                "r1": {
                    "upload": str(uuid4()),
                    "destination_path": "gncv://batch2/clientid2_R1.fastq.gz",
                    "last_status": {
                        "id": str(uuid4()),
                        "status": "assigned",
                    },
                }
            },
        },
    ],
)

MOCKED_EMPTY_UPLOADS = dict(
    meta=dict(next=None),
    results=[],
)


def test_run_prefix__bad_project_id(mocker):
    """Test run prefix when non-uuid string is used as project
    id."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
    )
    res = runner.invoke(
        run_prefix,
        [
            "1111111",
            "gncv://batch",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_not_called()
    assert "Project ID is not valid" in res.output


def test_run_prefix__bad_prefix(mocker):
    """Test run prefix when bad prefix string is used."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
    )
    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "prefix://path",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_not_called()
    assert "Prefix is not valid" in res.output


def test_run_prefix__bad_json(mocker):
    """Test run prefix when bad JSON is used."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
    )
    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "gncv://batch",
            "--metadata-json",
            "{bad:",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_not_called()
    assert "Metadata JSON is not valid" in res.output


def test_run_prefix__not_owned_project(mocker):
    """Test run prefix failure when project is not owned."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        side_effect=APIClientError(message="", status_code=403),
    )

    mocked_add_samples_to_project = mocker.patch.object(
        APIClient,
        "add_samples_to_project",
    )

    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "gncv://batch",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_called_once()
    mocked_add_samples_to_project.assert_not_called()
    assert "You do not have the sufficient permission" in res.output


def test_run_prefix__no_paths(mocker):
    """Test run prefix when no paths are present."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        return_value=SampleSheet(**MOCKED_EMPTY_UPLOADS),
    )
    mocked_add_samples_to_project = mocker.patch.object(
        APIClient,
        "add_samples_to_project",
        return_value=UploadSamples(**{}),
    )

    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "gncv://batch",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_called_once()
    mocked_add_samples_to_project.assert_not_called()
    assert "No matching paths found" in res.output


def test_run_prefix__assigning_samples_no_permission(mocker):
    """Test run prefix when permission is denied for assigning samples."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        return_value=SampleSheet(**MOCKED_UPLOADS),
    )
    mocked_add_samples_to_project = mocker.patch.object(
        APIClient,
        "add_samples_to_project",
        side_effect=APIClientError(message="", status_code=403),
    )

    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "gncv://batch",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_called_once()
    mocked_add_samples_to_project.assert_called_once()
    assert "You do not have the sufficient permission" in res.output


def test_run_prefix__assigning_samples_failed(mocker):
    """Test run prefix when assigning samples fails."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        return_value=SampleSheet(**MOCKED_UPLOADS),
    )
    mocked_add_samples_to_project = mocker.patch.object(
        APIClient,
        "add_samples_to_project",
        side_effect=APIClientError(message="", status_code=404),
    )

    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "gncv://batch",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_called_once()
    mocked_add_samples_to_project.assert_called_once()
    assert "You do not have the sufficient permission" not in res.output
    assert "There was an error assigning" in res.output


def test_run_prefix__success_with_json(mocker):
    """Test run prefix success with optional json passed."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        return_value=SampleSheet(**MOCKED_UPLOADS),
    )
    mocked_add_samples_to_project = mocker.patch.object(
        APIClient,
        "add_samples_to_project",
        return_value=UploadSamples(**{}),
    )

    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "gncv://batch",
            "--metadata-json",
            '{"somekey": "somevalue"}',
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_called_once()
    mocked_add_samples_to_project.assert_called_once()
    assert "Assigning metadata to the uploaded samples." in res.output


def test_run_prefix__filter_empty(mocker):
    """Test run prefix when all paths are filtered out."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        return_value=SampleSheet(**MOCKED_EMPTY_UPLOADS),
    )
    mocked_add_samples_to_project = mocker.patch.object(
        APIClient,
        "add_samples_to_project",
        return_value=UploadSamples(**{}),
    )

    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "gncv://batch",
            "--status",
            "assigned",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_called_once()
    mocked_add_samples_to_project.assert_not_called()
    assert "No matching paths found" in res.output


def test_run_prefix__filter_success(mocker):
    """Test run prefix when filter returns some paths."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_uploads_copy = copy.deepcopy(MOCKED_UPLOADS)
    del mocked_uploads_copy["results"][1]
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        return_value=SampleSheet(**mocked_uploads_copy),
    )
    mocked_add_samples_to_project = mocker.patch.object(
        APIClient,
        "add_samples_to_project",
    )

    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "gncv://batch",
            "--status",
            "unassigned",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_called_once()
    mocked_add_samples_to_project.assert_called_once()
    assert "Number of samples assigned to the project" in res.output
    assert "Assigning metadata to the uploaded samples." not in res.output


def test_run_prefix__success_without_last_status(mocker):
    """Test run prefix success."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_uploads_copy = copy.deepcopy(MOCKED_UPLOADS)
    del mocked_uploads_copy["results"][0]["fastq"]["r1"]["last_status"]
    del mocked_uploads_copy["results"][0]["fastq"]["r2"]["last_status"]
    del mocked_uploads_copy["results"][1]["fastq"]["r1"]["last_status"]
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        return_value=SampleSheet(**mocked_uploads_copy),
    )
    mocked_add_samples_to_project = mocker.patch.object(
        APIClient,
        "add_samples_to_project",
        return_value=UploadSamples(**{}),
    )

    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "gncv://batch",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_called_once()
    mocked_add_samples_to_project.assert_called_once()
    assert "Number of samples assigned to the project" in res.output
    assert "Assigning metadata to the uploaded samples." not in res.output


def test_run_prefix__success(mocker):
    """Test run prefix success."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        return_value=SampleSheet(**MOCKED_UPLOADS),
    )
    mocked_add_samples_to_project = mocker.patch.object(
        APIClient,
        "add_samples_to_project",
        return_value=UploadSamples(**{}),
    )

    res = runner.invoke(
        run_prefix,
        [
            str(uuid4()),
            "gncv://batch",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_called_once()
    mocked_add_samples_to_project.assert_called_once()
    assert "Number of samples assigned to the project" in res.output
    assert "Assigning metadata to the uploaded samples." not in res.output



"""Describe constants in samples subcommand."""
from typing import Optional

from gencove.constants import Optionals


# pylint: disable=too-few-public-methods
class SamplesOptions(Optionals):
    """SamplesOptions model"""

    status: Optional[str]
    archive_status: Optional[str]
    search: Optional[str]



"""Describe constants in samples subcommand."""
from typing import Optional

from gencove.constants import Optionals


# pylint: disable=too-few-public-methods
class UploadsOptions(Optionals):
    """UploadsOptions model"""

    status: Optional[str]
    search: Optional[str]



"""VCR filters for projects tests."""
import json
from urllib.parse import urlparse

from gencove.tests.decorators import parse_response_to_json
from gencove.tests.filters import _replace_uuid_from_url
from gencove.tests.utils import MOCK_UUID


@parse_response_to_json
def filter_list_projects_response(response, json_response):
    """Filter list project sensitive data from response."""
    if "results" in json_response:
        for result in json_response["results"]:
            if "id" in result:
                result["id"] = MOCK_UUID
            if "name" in result:
                result["name"] = "mock name"
            if "description" in result and result["description"]:
                result["description"] = "mock description"
            if "organization" in result:
                result["organization"] = MOCK_UUID
            if "sample_count" in result:
                result["sample_count"] = 1
            if "pipeline_capabilities" in result:
                result["pipeline_capabilities"] = MOCK_UUID
            if "roles" in result:
                result["roles"]["organization"]["id"] = MOCK_UUID
            if "webhook_url" in result and result["webhook_url"]:
                result["webhook_url"] = "http://mockurl.com"
    return response, json_response


def filter_pipeline_capabilities_request(request):
    """Filter pipeline capabilities sensitive data from request."""
    return _replace_uuid_from_url(request, "pipeline-capabilities")


@parse_response_to_json
def filter_pipeline_capabilities_response(response, json_response):
    """Filter pipeline capabilities sensitive data from response."""
    if "id" in json_response:
        json_response["id"] = MOCK_UUID
    if "name" in json_response:
        json_response["name"] = "mock name"
    return response, json_response


def filter_get_project_samples_request(request):
    """Filter project samples sensitive data from request."""
    return _replace_uuid_from_url(request, "project-samples")


@parse_response_to_json
def filter_get_project_samples_response(response, json_response):
    """Filter list project samples sensitive data from response."""
    if "results" in json_response:
        for result in json_response["results"]:
            if "id" in result:
                result["id"] = MOCK_UUID
            if "client_id" in result:
                result["client_id"] = "mock client_id"
            if "physical_id" in result and result["physical_id"]:
                result["physical_id"] = "mock physical_id"
            if "legacy_id" in result and result["legacy_id"]:
                result["legacy_id"] = "mock legacy_id"
            if "last_status" in result:
                result["last_status"]["id"] = MOCK_UUID
                result["last_status"]["status"] = "mock status"
                if result["last_status"]["note"]:
                    result["last_status"]["note"] = "mock note"
            if "archive_last_status" in result:
                result["archive_last_status"]["id"] = MOCK_UUID
                result["archive_last_status"]["status"] = "mock status"
                if result["archive_last_status"]["transition_cutoff"]:
                    result["archive_last_status"][
                        "transition_cutoff"
                    ] = "mock transition_cutoff"
    return response, json_response


def filter_get_project_batch_types_request(request):
    """Filter project samples sensitive data from request."""
    return _replace_uuid_from_url(request, "project-batch-types")


def filter_post_project_restore_samples(request):
    """Filter project restore samples sensitive data from request."""
    if "project-restore-samples" in request.path:
        request = _replace_uuid_from_url(request, "project-restore-samples")
        try:
            body = json.loads(request.body)
            samples = [MOCK_UUID for _ in body["sample_ids"]]
            body["sample_ids"] = samples
            request.body = json.dumps(body).encode()
        except json.decoder.JSONDecodeError:
            pass
    return request


def filter_project_batches_request(request):
    """Filter project batches sensitive data from request."""
    if "project-batches" in request.path:
        request = _replace_uuid_from_url(request, "project-batches")
        try:
            body = json.loads(request.body)
            body["name"] = "Mock name"
            body["batch_type"] = "Mock batch_type"
            body["sample_ids"] = [MOCK_UUID for _ in body["sample_ids"]]
            request.body = json.dumps(body).encode()
        except (json.decoder.JSONDecodeError, TypeError):
            pass
    return request


@parse_response_to_json
def filter_project_batches_response(response, json_response):
    """Filter project batches sensitive data from response."""
    if "results" in json_response:
        for result in json_response["results"]:
            if "id" in result:
                result["id"] = MOCK_UUID
            if "name" in result:
                result["name"] = "Mock name"
            if "batch_type" in result:
                result["batch_type"] = "Mock batch_type"
            if "last_status" in result:
                result["last_status"]["id"] = MOCK_UUID
            for file in result.get("files", []):
                if "id" in file:
                    file["id"] = MOCK_UUID
                if "s3_path" in file:
                    file["s3_path"] = "mock/file.zip"
                if "size" in file and file["size"]:
                    file["size"] = "1"
                if "download_url" in file:
                    file["download_url"] = "https://example.com/file.zip"
    return response, json_response


def filter_batches_request(request):
    """Filter batches sensitive data from request."""
    return _replace_uuid_from_url(request, "batches")


@parse_response_to_json
def filter_batches_response(response, json_response):
    """Filter batches sensitive data from response."""
    if "id" in json_response:
        json_response["id"] = MOCK_UUID
    if "name" in json_response:
        json_response["name"] = "mock name"
    if "batch_type" in json_response:
        json_response["batch_type"] = "Mock batch_type"
    if "sample_ids" in json_response:
        json_response["sample_ids"] = [
            MOCK_UUID for _ in json_response["sample_ids"]
        ]
    if "last_status" in json_response:
        json_response["last_status"]["id"] = MOCK_UUID
    for file in json_response.get("files", []):
        if "id" in file:
            file["id"] = MOCK_UUID
        if "s3_path" in file:
            file["s3_path"] = f"mock/{file['s3_path'].split('/')[-1]}"
        if "size" in file and file["size"]:
            file["size"] = "1"
        if "download_url" in file:
            filename = urlparse(file["download_url"]).path.split("/")[-1]
            file["download_url"] = f"https://example.com/{filename}"
    return response, json_response



"""AutoImport S3 URI to a project subcommand."""

import json

from ..... import client
from .....exceptions import ValidationError
from ....base import Command  # noqa: I100
from ....utils import is_valid_json, is_valid_uuid


class S3AutoImport(Command):
    """S3AutoImport autoimport command executor."""

    def __init__(self, project_id, s3_uri, credentials, options):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.s3_uri = s3_uri
        self.metadata_json = options.metadata_json

    def initialize(self):
        """Initialize s3-autoimport subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if is_valid_uuid(self.project_id) is False:
            raise ValidationError("Project ID is not valid. Exiting.")
        if self.metadata_json and is_valid_json(self.metadata_json) is False:
            raise ValidationError("Metadata JSON is not valid. Exiting.")

    # no retry for timeouts in order to avoid duplicate heavy operations on
    # the backend
    def execute(self):
        """Make a request to create a job to automatically import
        from S3 URI to a Gencove project.
        """
        self.echo_debug(
            "Create AutoImport job from "
            "s3_uri {} to project {}".format(self.s3_uri, self.project_id)
        )

        try:
            # prepare metadata
            metadata = None
            if self.metadata_json is not None:
                metadata = json.loads(self.metadata_json)
                self.echo_info(
                    "Metadata will be assigned to the imported Biosamples."
                )
            autoimport_from_s3 = self.api_client.autoimport_from_s3(
                project_id=self.project_id,
                s3_uri=self.s3_uri,
                metadata=metadata,
            )
            self.echo_info(
                "Request to create a periodic import job from"
                "S3 URI to project accepted."
            )
            self.echo_data(
                "\t".join(
                    [
                        str(autoimport_from_s3.id),
                        autoimport_from_s3.topic_arn,
                    ]
                )
            )
        except client.APIClientError:
            self.echo_error(
                "There was an error creating an import job of S3 URI."
            )
            raise



"""Project batch get shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import GetBatch


@click.command("get-batch")
@click.argument("batch_id")
@click.option(
    "--output-filename",
    help="Output filename for batch deliverable.",
    type=click.Path(),
    required=False,
    default=None,
)
@add_options(common_options)
@click.option(
    "--no-progress",
    is_flag=True,
    help="If specified, no progress bar is shown.",
)
# pylint: disable=too-many-arguments
def get_batch(
    batch_id, output_filename, host, email, password, api_key, no_progress
):
    """Get batch that is available for a project."""
    GetBatch(
        batch_id,
        output_filename,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
        no_progress,
    ).run()



"""General filters for VCR cassettes."""
import copy
import re
from urllib.parse import urlparse, urlunparse

from gencove.tests.decorators import parse_response_to_json
from gencove.tests.utils import MOCK_UUID


@parse_response_to_json
def filter_jwt(respose, response_json):
    """Replace the refresh and access token with some mock up data."""
    if "refresh" in response_json and "access" in response_json:
        response_json["refresh"] = "mock_refresh"
        response_json["access"] = "mock_access"
    return respose, response_json


def replace_gencove_url_vcr(request):
    """Replace urls with 'gencove' in it."""
    request = copy.deepcopy(request)
    if "gencove" in request.uri:
        request.uri = urlunparse(
            urlparse(request.uri)._replace(netloc="www.example.com")
        )
    return request


def _replace_uuid_from_url(request, endpoint):
    """Removes the id from the last part of the URL."""
    request = copy.deepcopy(request)
    if endpoint in request.path:
        uuid = re.compile(
            r"[0-9A-F]{8}-[0-9A-F]{4}-[4][0-9A-F]{3}-[89AB][0-9A-F]{3}-[0-9A-F]{12}",  # noqa: E501 pylint: disable=line-too-long
            re.IGNORECASE,
        )
        request.uri = uuid.sub(MOCK_UUID, request.uri)
    return request


def filter_aws_headers(response):
    """Removes the amz id, request-id and version-id from the response.
    This can't be done in filter_headers.
    """
    for key in ["x-amz-id-2", "x-amz-request-id", "x-amz-version-id"]:
        if key in response["headers"]:
            response["headers"][key] = [f"mock_{key}"]
    return response


def replace_s3_from_url(request):
    """Mock the S3 URLS."""
    request = copy.deepcopy(request)
    if "s3.amazonaws.com" in request.uri:
        filename = urlparse(request.uri).path.split("/")[-1]
        request.uri = f"https://s3.amazonaws.com/{filename}"
    return request


def mock_binary_response(response):
    """Replace binary responses with a placeholder"""
    is_binary = "binary/octet-stream" in response["headers"].get(
        "Content-Type", []
    )
    if is_binary:
        response["body"]["string"] = b"""QUFBQkJC"""
        if "Content-Disposition" in response["headers"]:
            response["headers"]["Content-Disposition"] = ["attachment;"]
        if "Content-Length" in response["headers"]:
            response["headers"]["Content-Length"] = ["123"]
    return response


def filter_samples_request(request):
    """Filter samples sensitive data from request."""
    return _replace_uuid_from_url(request, "samples")


def _filter_sample(result):
    """Common function that filters sample sensitive data."""
    if "id" in result:
        result["id"] = MOCK_UUID
    if "client_id" in result:
        result["client_id"] = "mock_client_id"
    if "project" in result:
        result["project"] = MOCK_UUID
    if "physical_id" in result and result["physical_id"]:
        result["physical_id"] = "mock_physical_id"
    if "legacy_id" in result and result["legacy_id"]:
        result["legacy_id"] = "mock_legacy_id"
    if "last_status" in result:
        result["last_status"]["id"] = MOCK_UUID
    if "archive_last_status" in result:
        result["archive_last_status"]["id"] = MOCK_UUID
    for file in result.get("files", []):
        if "id" in file:
            file["id"] = MOCK_UUID
        if "s3_path" in file:
            file["s3_path"] = f"mock/{file['s3_path'].split('/')[-1]}"
        if "size" in file and file["size"]:
            file["size"] = "1"
        if "download_url" in file and file["download_url"]:
            filename = urlparse(file["download_url"]).path.split("/")[-1]
            file["download_url"] = f"https://example.com/{filename}"


@parse_response_to_json
def filter_project_samples_response(response, response_json):
    """Filter project samples sensitive data from response."""
    for result in response_json.get("results", []):
        _filter_sample(result)
    return response, response_json


@parse_response_to_json
def filter_samples_response(response, response_json):
    """Filter samples sensitive data from response."""
    _filter_sample(response_json)
    return response, response_json



"""Tests upload command of Gencove CLI."""

# pylint: disable=too-many-lines, import-error

import json
import operator
import os
from uuid import uuid4

from click.testing import CliRunner


from gencove.cli import upload
from gencove.client import APIClient, APIClientError, APIClientTimeout
from gencove.command.upload.utils import upload_file
from gencove.constants import ApiEndpoints, UPLOAD_PREFIX
from gencove.models import SampleSheet, UploadSamples, UploadsPostData
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import (
    filter_aws_headers,
    filter_jwt,
    replace_gencove_url_vcr,
)
from gencove.tests.upload.vcr.filters import (
    filter_project_samples_request,
    filter_project_samples_response,
    filter_sample_sheet_response,
    filter_upload_credentials_response,
    filter_upload_post_data_response,
    filter_upload_request,
    filter_volatile_dates,
)
from gencove.tests.utils import get_vcr_response
from gencove.utils import get_regular_progress_bar, get_s3_client_refreshable

import pytest  # pylint: disable=wrong-import-order

from vcr import VCR  # pylint: disable=wrong-import-order


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/upload/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
            "X-Amz-Security-Token",
            "X-Amz-Date",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "filter_query_parameters": [
            ("search", "gncv://cli-mock/test.fastq.gz")
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_upload_request,
            filter_project_samples_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_upload_credentials_response,
            filter_upload_post_data_response,
            filter_volatile_dates,
            filter_aws_headers,
            filter_sample_sheet_response,
            filter_project_samples_response,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_upload(credentials, vcr, recording, mocker):
    """Sanity check that upload is ok."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")
        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            response = get_vcr_response("/api/v2/uploads-post-data/", vcr)
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**response),
            )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )

        res = runner.invoke(
            upload,
            ["cli_test_data", *credentials],
        )
        assert not res.exception
        assert res.exit_code == 0
        assert (
            "Uploading cli_test_data/test.fastq.gz to gncv://" in res.output
        )
        assert "All files were successfully uploaded." in res.output
        mocked_get_credentials.assert_called_once()
        if not recording:
            mocked_get_upload_details.assert_called_once()
        mocked_upload_file.assert_called_once()
        assert not mocked_upload_file.call_args[1]["no_progress"]


@pytest.mark.vcr
@assert_authorization
def test_upload_no_files_found(
    credentials, mocker, recording, using_api_key, vcr
):  # pylint: disable=unused-argument
    """Test that no fastq files found exits upload."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            upload,
            ["cli_test_data", *credentials],
        )
        assert res.exit_code == 1
        assert "No FASTQ files found in the path" in res.output
        if using_api_key:
            assert vcr.play_count == 0
        elif not recording:
            # Check that we are making just one request (to the login) if we
            # are not recording the cassette
            assert vcr.play_count == 1


@pytest.mark.vcr
@assert_authorization
def test_upload_invalid_destination(
    credentials, mocker, recording, using_api_key, vcr
):  # pylint: disable=unused-argument
    """Test that invalid destination exists upload."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")
        res = runner.invoke(
            upload,
            ["cli_test_data", "foobar_dir", *credentials],
        )
        assert res.exit_code == 1
        assert (
            "Invalid destination path. Must start with '{}'".format(
                UPLOAD_PREFIX
            )
            in res.output
        )
        if using_api_key:
            assert vcr.play_count == 0
        elif not recording:
            # Check that we are making just one request (to the login) if we
            # are playing the cassette
            assert vcr.play_count == 1


@pytest.mark.vcr
@assert_authorization
def test_upload_project_id_not_uuid(
    credentials, mocker, recording, using_api_key, vcr
):  # pylint: disable=unused-argument
    """Test that project id is valid UUID when uploading."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")
        res = runner.invoke(
            upload,
            ["cli_test_data", "--run-project-id", "1234", *credentials],
        )

        assert res.exit_code == 1
        assert "--run-project-id is not valid" in res.output
        if using_api_key:
            assert vcr.play_count == 0
        elif not recording:
            # Check that we are making just one request (to the login) if we
            # are playing the cassette
            assert vcr.play_count == 1


@pytest.mark.vcr(
    filter_query_parameters=[("search", "gncv://cli-mock/test.fastq.gz")],
)
@assert_authorization
def test_upload_and_run_immediately(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )
        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response(
                "/api/v2/sample-sheet/", vcr
            )
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )
        res = runner.invoke(
            upload,
            ["cli_test_data", "--run-project-id", project_id, *credentials],
        )

        assert not res.exception
        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        if not recording:
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()
        mocked_regular_progress_bar.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately__with_metadata(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )

        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )

        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response(
                "/api/v2/sample-sheet/", vcr
            )
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                "--run-project-id",
                project_id,
                "--metadata",
                "[1,2]",
                *credentials,
            ],
        )

        assert not res.exception
        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        mocked_regular_progress_bar.assert_called_once()
        if not recording:
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately__invalid_metadata(
    credentials, mocker, project_id
):
    """Upload and assign right away with invalid metadata."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable"
        )

        mocked_get_upload_details = mocker.patch.object(
            APIClient,
            "get_upload_details",
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file"
        )
        mocked_get_sample_sheet = mocker.patch.object(
            APIClient,
            "get_sample_sheet",
        )
        mocked_assign_sample = mocker.patch.object(
            APIClient, "add_samples_to_project"
        )

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--metadata",
                "[1,2,3",
            ],
        )

        assert res.exit_code == 1
        mocked_get_credentials.assert_not_called()
        mocked_get_upload_details.assert_not_called()
        mocked_upload_file.assert_not_called()
        mocked_get_sample_sheet.assert_not_called()
        mocked_assign_sample.assert_not_called()
        assert "--metadata is not valid JSON" in res.output


@pytest.mark.vcr
@assert_authorization
def test_upload__with_metadata_without_project_id(credentials, mocker):
    """Test that project id is present if attaching metadata when
    uploading."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable"
        )
        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--metadata",
                "[1,2]",
            ],
        )

        assert res.exit_code == 1
        assert (
            "--metadata cannot be used without --run-project-id" in res.output
        )
        mocked_get_credentials.assert_not_called()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_something_went_wrong(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away did't work."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )
        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
        mocked_get_sample_sheet = mocker.patch.object(
            APIClient,
            "get_sample_sheet",
            return_value=SampleSheet(
                **{"meta": {"next": None}, "results": []}
            ),
        )
        mocked_assign_sample = mocker.patch.object(
            APIClient, "add_samples_to_project"
        )

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
            ],
        )

        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        if not recording:
            mocked_get_upload_details.assert_called_once()
        mocked_upload_file.assert_called_once()
        mocked_get_sample_sheet.assert_called()
        mocked_assign_sample.assert_not_called()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_with_output_to_file(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away, then save results to a file."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )

        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )

        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response(
                "/api/v2/sample-sheet/", vcr
            )
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--output",
                "samples.json",
            ],
        )

        assert not res.exception
        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        mocked_regular_progress_bar.assert_called_once()

        if not recording:
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()
            with open("samples.json", "r") as output_file:
                output_content = output_file.read()
            # r2 fastq key is not present on the file
            assert (
                project_sample_response["uploads"][0]["fastq"]["r2"] is None
            )
            del project_sample_response["uploads"][0]["fastq"]["r2"]
            assert (
                json.dumps(project_sample_response["uploads"], indent=4)
                == output_content
            )


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_with_output_to_nested_file(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away, then save results to a file."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )

        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )

        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response(
                "/api/v2/sample-sheet/", vcr
            )
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--output",
                "somefolder/samples.json",
            ],
        )

        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        mocked_regular_progress_bar.assert_called_once()

        if not recording:
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()
            with open("somefolder/samples.json", "r") as output_file:
                output_content = output_file.read()
            # r2 fastq key is not present on the file
            assert (
                project_sample_response["uploads"][0]["fastq"]["r2"] is None
            )
            del project_sample_response["uploads"][0]["fastq"]["r2"]
            assert (
                json.dumps(project_sample_response["uploads"], indent=4)
                == output_content
            )


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_with_stdout(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away, then print out the results."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )

        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response(
                "/api/v2/sample-sheet/", vcr
            )
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )
            # These mocks needs to be here and without side effect otherwise
            # a weird behavior on the stdout will happend and the test fails.
            mocked_upload_file = mocker.patch(
                "gencove.command.upload.main.upload_file"
            )

            mocker.patch(
                "gencove.command.upload.main.get_regular_progress_bar",
            )

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--output",
                "-",
            ],
        )

        assert res.exit_code == 0

        mocked_get_credentials.assert_called_once()
        if not recording:
            mocked_upload_file.assert_called_once()
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()
            # r2 fastq key is not present on the file
            assert (
                project_sample_response["uploads"][0]["fastq"]["r2"] is None
            )
            del project_sample_response["uploads"][0]["fastq"]["r2"]
            assert (
                json.dumps(project_sample_response["uploads"], indent=4)
                in res.output
            )


@pytest.mark.vcr
@assert_authorization
def test_upload_without_progressbar(credentials, mocker, recording, vcr):
    """Upload do not show the progress bar."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file"
        )
        if not recording:
            # Mock get_upload_details only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--no-progress",
            ],
        )

        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        assert mocked_upload_file.call_args[1]["no_progress"]
        if not recording:
            mocked_get_upload_details.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_without_progressbar(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away."""
    # pylint: disable=too-many-locals
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )

        mocked_regular_progress_bar = mocker.patch(
            "gencove.command.upload.main.get_regular_progress_bar",
            side_effect=get_regular_progress_bar,
        )
        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
            sample_sheet_response = get_vcr_response(
                "/api/v2/sample-sheet/", vcr
            )
            mocked_get_sample_sheet = mocker.patch.object(
                APIClient,
                "get_sample_sheet",
                return_value=SampleSheet(**sample_sheet_response),
            )
            project_sample_response = get_vcr_response(
                "/api/v2/project-samples/", vcr, operator.contains
            )
            mocked_assign_sample = mocker.patch.object(
                APIClient,
                "add_samples_to_project",
                return_value=UploadSamples(**project_sample_response),
            )

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--no-progress",
            ],
        )

        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        mocked_regular_progress_bar.assert_not_called()

        if not recording:
            mocked_get_upload_details.assert_called_once()
            mocked_get_sample_sheet.assert_called()
            mocked_assign_sample.assert_called_once()


@pytest.mark.vcr
@assert_authorization
def test_upload_and_run_immediately_slow_response_retry(
    credentials, mocker, project_id, recording, vcr
):
    """Upload and assign right away and retry on slow response."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable",
            side_effect=get_s3_client_refreshable,
        )
        mocked_upload_file = mocker.patch(
            "gencove.command.upload.main.upload_file", side_effect=upload_file
        )
        if not recording:
            # Mock get_upload credentials only if using the cassettes, since
            # we mock the return value.
            upload_details_response = get_vcr_response(
                "/api/v2/uploads-post-data/", vcr
            )
            mocked_get_upload_details = mocker.patch.object(
                APIClient,
                "get_upload_details",
                return_value=UploadsPostData(**upload_details_response),
            )
        mocked_get_sample_sheet = mocker.patch.object(
            APIClient,
            "get_sample_sheet",
            side_effect=APIClientTimeout(
                "Could not connect to the api server"
            ),
        )

        res = runner.invoke(
            upload,
            [
                "cli_test_data",
                *credentials,
                "--run-project-id",
                project_id,
                "--no-progress",
            ],
        )

        assert res.exit_code == 0
        mocked_get_credentials.assert_called_once()
        mocked_upload_file.assert_called_once()
        if not recording:
            mocked_get_upload_details.assert_called_once()
        assert mocked_get_sample_sheet.call_count == 5
        assert "there was an error automatically running them" in res.output


def test_upload_retry_after_unauthorized(mocker):
    """Test that the upload is performed when retrying due to unauthorized
    http response.
    """
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        with open("cli_test_data/test.fastq.gz", "w") as fastq_file:
            fastq_file.write("AAABBB")

        def _login(
            self, email, password, otp_token=None
        ):  # pylint: disable=unused-argument
            # pylint: disable=protected-access
            self._jwt_refresh_token = "refresh_token"

        mocked_login = mocker.patch.object(
            APIClient, "login", side_effect=_login, autospec=True
        )

        force_refresh_jwt = True

        def _request(
            endpoint,
            *args,  # pylint: disable=unused-argument
            **kwargs,  # pylint: disable=unused-argument
        ):
            if ApiEndpoints.UPLOAD_DETAILS.value == endpoint:
                nonlocal force_refresh_jwt
                if force_refresh_jwt:
                    force_refresh_jwt = False
                    raise APIClientError("Test error.", 401)
                return {
                    "id": str(uuid4()),
                    "last_status": {"id": str(uuid4()), "status": ""},
                    "s3": {"bucket": "test", "object_name": "test"},
                }
            if ApiEndpoints.REFRESH_JWT.value == endpoint:
                return {"access": ""}
            return {}

        mocked_request = mocker.patch.object(
            APIClient, "_request", side_effect=_request
        )

        mocked_get_credentials = mocker.patch(
            "gencove.command.upload.main.get_s3_client_refreshable"
        )

        res = runner.invoke(
            upload,
            ["cli_test_data"],
            input="\n".join(["foo@bar.com", "123456"]),
        )

        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_get_credentials.assert_called_once()
        # Call count = upload details + refresh jwt + retry upload details
        assert mocked_request.call_count == 3
        assert force_refresh_jwt is False



"""Entry point into upload command."""
import json
import os
import uuid
from datetime import datetime
from time import sleep

import backoff

from gencove.client import (  # noqa: I100
    APIClientError,
    APIClientTimeout,
    APIClientTooManyRequestsError,
    CustomEncoder,
)
from gencove.command.base import Command
from gencove.command.utils import is_valid_uuid
from gencove.constants import (
    FASTQ_MAP_EXTENSION,
    SampleAssignmentStatus,
    UPLOAD_PREFIX,
)
from gencove.exceptions import ValidationError
from gencove.utils import (
    batchify,
    get_regular_progress_bar,
    get_s3_client_refreshable,
)

from .constants import (
    ASSIGN_ERROR,
    FASTQ_EXTENSIONS,
    TMP_UPLOADS_WARNING,
    UploadStatuses,
)
from .exceptions import SampleSheetError, UploadError, UploadNotFound
from .multi_file_reader import MultiFileReader
from .utils import (
    get_filename_from_path,
    get_get_upload_details_retry_predicate,
    get_gncv_path,
    parse_fastqs_map_file,
    seek_files_to_upload,
    upload_file,
    upload_multi_file,
)
from ..utils import is_valid_json
from ...constants import ASSIGN_BATCH_SIZE


# pylint: disable=too-many-instance-attributes
class Upload(Command):
    """Upload command executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self, source, destination, credentials, options, output, no_progress
    ):
        super().__init__(credentials, options)
        self.source = source
        self.destination = destination
        self.project_id = options.project_id
        self.fastqs = []
        self.fastqs_map = {}
        self.upload_ids = set()
        self.output = output
        self.assigned_samples = []
        self.no_progress = no_progress
        self.metadata = options.metadata

    @staticmethod
    def generate_gncv_destination():
        """Autogenerate gencove destination path."""
        return "{}cli-{}-{}".format(
            UPLOAD_PREFIX,
            datetime.utcnow().strftime("%Y%m%d%H%M%S"),
            uuid.uuid4().hex,
        )

    def initialize(self):
        """Initialize upload command parameters from provided arguments."""
        self.echo_debug("Host is {}".format(self.options.host))
        self.echo_warning(TMP_UPLOADS_WARNING)

        # fmt: off
        if os.path.isfile(self.source) and self.source.endswith(FASTQ_MAP_EXTENSION):  # noqa: E501
            self.echo_debug("Scanning fastqs map file")
            self.fastqs_map = parse_fastqs_map_file(self.source)
            self.echo_debug("got fastq pairs: {}".format(self.fastqs_map))
        else:
            self.echo_debug("Seeking files to upload")
            self.fastqs = list(seek_files_to_upload(self.source))

        if not self.destination:
            self.destination = self.generate_gncv_destination()
            self.echo_info(
                "Files will be uploaded to: {}".format(self.destination)
            )
        # fmt: on

        # Make sure there is just one trailing slash. Only exception is
        # UPLOAD_PREFIX itself, which can have two trailing slashes.
        if self.destination != UPLOAD_PREFIX:
            self.destination = self.destination.rstrip("/")
            self.destination += "/"

        if self.is_credentials_valid:
            self.login()

    def validate(self):
        """Validate command setup before execution.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        # fmt: off
        if self.destination and not self.destination.startswith(UPLOAD_PREFIX):
            self.echo_error(
                "Invalid destination path. Must start with '{}'".format(
                    UPLOAD_PREFIX
                )
            )
            raise ValidationError("Bad configuration. Exiting.")
        # fmt: on

        if not self.fastqs and not self.fastqs_map:
            self.echo_error(
                "No FASTQ files found in the path. "
                "Only following files are accepted: {}".format(
                    FASTQ_EXTENSIONS
                )
            )
            raise ValidationError("Bad configuration. Exiting.")

        if self.output and not self.project_id:
            raise ValidationError(
                "--output cannot be used without --run-project-id"
            )

        if self.metadata and not self.project_id:
            raise ValidationError(
                "--metadata cannot be used without --run-project-id"
            )

        if self.project_id and is_valid_uuid(self.project_id) is False:
            raise ValidationError("--run-project-id is not valid. Exiting.")

        # validate metadata json
        if self.metadata and is_valid_json(self.metadata) is False:
            raise ValidationError("--metadata is not valid JSON. Exiting.")

    def execute(self):
        """Upload fastq files from host system to Gencove cloud.

        If project id was provided, all fastq files will
        be assigned to this project, after upload.
        """
        s3_client = get_s3_client_refreshable(
            self.api_client.get_upload_credentials
        )
        try:
            if self.fastqs:
                self.upload_from_source(s3_client)
            elif self.fastqs_map:
                self.upload_from_map_file(s3_client)
        except UploadError:
            return

        self.echo_debug("Upload ids are now: {}".format(self.upload_ids))
        if self.project_id:
            self.echo_debug("Cooling down period.")
            sleep(10)
            self.assign_uploads_to_project()
            if self.output:
                self.output_list()

    def upload_from_source(self, s3_client):
        """Upload command with <source> argument provided."""
        for file_path in self.fastqs:
            upload = self.upload_from_file_path(file_path, s3_client)
            if self.project_id and upload:
                self.upload_ids.add(upload.id)

        self.echo_info("All files were successfully uploaded.")

    def upload_from_map_file(self, s3_client):
        """Upload fastq files from a csv file."""
        for key, fastqs in self.fastqs_map.items():
            upload = self.concatenate_and_upload_fastqs(
                key, fastqs, s3_client
            )
            if self.project_id and upload:
                self.upload_ids.add(upload.id)

        self.echo_info("All files were successfully uploaded.")

    def concatenate_and_upload_fastqs(self, key, fastqs, s3_client):
        """Upload fastqs parts as one file."""
        client_id, r_notation = key
        self.echo_debug(
            "Uploading fastq. client_id={} r_notation={}".format(
                client_id, r_notation
            )
        )
        self.echo_debug("FASTQS: {}".format(fastqs))

        gncv_path = self.destination + get_gncv_path(client_id, r_notation)
        self.echo_debug("Calculated gncv path: {}".format(gncv_path))

        upload_details = self.get_upload_details(gncv_path)

        if (
            upload_details.last_status
            and upload_details.last_status.status == UploadStatuses.DONE.value
        ):
            self.echo_info("File was already uploaded: {}".format(gncv_path))
            return upload_details

        self.echo_info("Uploading to {}".format(gncv_path))
        upload_multi_file(
            s3_client,
            MultiFileReader(fastqs),
            upload_details.s3.bucket,
            upload_details.s3.object_name,
            self.no_progress,
        )
        return upload_details

    def upload_from_file_path(self, file_path, s3_client):
        """Prepare file and upload, if it wasn't uploaded yet.

        Args:
            file_path (str): a local system path to a file to be uploaded.
            s3_client (boto3 s3 client): instantiated boto3 S3 client.

        Returns:
            dict representing upload details
        """
        clean_file_path = get_filename_from_path(file_path, self.source)
        self.echo_debug(
            "Uploading clean file path: {}".format(clean_file_path)
        )
        gncv_notated_path = self.destination + clean_file_path

        self.echo_info(
            "Checking if file was already uploaded: {}".format(
                clean_file_path
            )
        )

        try:
            upload_details = self.get_upload_details(gncv_notated_path)
        except APIClientError as err:
            if err.status_code == 400:
                self.echo_info(err.message)
                raise UploadError  # pylint: disable=W0707
            raise err

        if (
            upload_details.last_status
            and upload_details.last_status.status == UploadStatuses.DONE.value
        ):
            self.echo_info(
                "File was already uploaded: {}".format(clean_file_path)
            )
            return upload_details

        self.echo_info(
            "Uploading {} to {}".format(file_path, gncv_notated_path)
        )
        upload_file(
            s3_client=s3_client,
            file_name=file_path,
            bucket=upload_details.s3.bucket,
            object_name=upload_details.s3.object_name,
            no_progress=self.no_progress,
        )
        return upload_details

    @backoff.on_predicate(
        backoff.expo, get_get_upload_details_retry_predicate, max_tries=10
    )
    @backoff.on_exception(
        backoff.expo, APIClientTooManyRequestsError, max_tries=20
    )
    def get_upload_details(self, gncv_path):
        """Get upload details with retry for last status update."""
        return self.api_client.get_upload_details(gncv_path)

    def assign_uploads_to_project(self):
        """Assign uploads to a project and trigger a run."""
        self.echo_info(
            "Assigning uploads to project {}".format(self.project_id)
        )

        try:
            samples = self.build_samples(self.upload_ids)
        except (UploadError, SampleSheetError, UploadNotFound):
            self.echo_warning(
                ASSIGN_ERROR.format(self.project_id, self.destination)
            )
            return

        if not samples:
            self.echo_debug("No related samples were found")
            self.echo_warning(
                ASSIGN_ERROR.format(self.project_id, self.destination)
            )
            return

        self.echo_debug("Sample sheet now is: {}".format(samples))

        self.echo_debug(
            "Assigning samples to project ({})".format(self.project_id)
        )

        assigned_count = 0
        if not self.no_progress:
            progress_bar = get_regular_progress_bar(
                len(samples), "Assigning: "
            )
            progress_bar.start()
        for samples_batch in batchify(samples, batch_size=ASSIGN_BATCH_SIZE):
            try:
                samples_batch_len = len(samples_batch)
                self.echo_debug(
                    "Assigning batch: {}".format(samples_batch_len)
                )
                metadata_api = None
                if self.metadata is not None:
                    metadata_api = json.loads(self.metadata)
                assigned_batch = self.api_client.add_samples_to_project(
                    samples_batch, self.project_id, metadata_api
                )
                if assigned_batch.uploads:
                    self.assigned_samples.extend(assigned_batch.uploads)
                assigned_count += samples_batch_len
                if not self.no_progress:
                    progress_bar.update(samples_batch_len)
                self.echo_debug("Total assigned: {}".format(assigned_count))
            except APIClientError as err:
                self.echo_debug(err)
                self.echo_warning(
                    "There was an error assigning/running samples."
                )
                if assigned_count > 0:
                    self.echo_warning(
                        "Some of the samples were assigned. "
                        "Please use the Web UI to assign "
                        "the rest of the samples"
                    )
                else:
                    self.echo_warning(
                        ASSIGN_ERROR.format(self.project_id, self.destination)
                    )
                if not self.no_progress:
                    progress_bar.finish()
                return
        if not self.no_progress:
            progress_bar.finish()
        self.echo_info("Assigned all samples to a project")

    @backoff.on_exception(
        backoff.expo, (SampleSheetError, UploadNotFound), max_time=300
    )
    def build_samples(self, uploads):
        """Get samples for current uploads.

        Returns:
            list of dict: a list of samples for the uploads.
        """
        # make a copy of uploads so as not to change the input
        search_uploads = uploads.copy()
        samples = []
        for sample_sheet in self.sample_sheet_paginator():
            if not sample_sheet:
                self.echo_debug("Sample sheet returned empty.")
                raise UploadError

            for sample in sample_sheet:
                self.echo_debug("Checking sample: {}".format(sample))
                add_it = False
                if sample.fastq and sample.fastq.r1:
                    if sample.fastq.r1.upload in search_uploads:
                        add_it = True
                        search_uploads.remove(sample.fastq.r1.upload)
                        self.echo_debug(
                            "Found sample for upload r1: {}".format(
                                sample.fastq.r1.upload
                            )
                        )
                    else:
                        self.echo_debug(
                            "R1 upload not found. sample {} uploads {}".format(
                                sample, search_uploads
                            )
                        )
                        raise UploadNotFound
                if sample.fastq and sample.fastq.r2:
                    if sample.fastq.r2.upload in search_uploads:
                        add_it = True
                        search_uploads.remove(sample.fastq.r2.upload)
                        self.echo_debug(
                            "Found sample for upload r2: {}".format(
                                sample.fastq.r2.upload
                            )
                        )
                    else:
                        self.echo_debug(
                            "R2 upload not found. sample {} uploads {}".format(
                                sample, search_uploads
                            )
                        )
                        raise UploadNotFound

                if add_it:
                    samples.append(sample)

        if search_uploads:
            self.echo_debug(
                "Have uploads without samples: {}".format(search_uploads)
            )
            raise SampleSheetError

        return samples

    # duplicated in projects/run_prefix/main.py with the exception difference
    def sample_sheet_paginator(self):
        """Paginate over all sample sheets for the destination.

        Yields:
            paginated lists of samples
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get sample sheet page")
            try:
                resp = self.get_sample_sheet(next_link)
                yield resp.results
                next_link = resp.meta.next
                more = next_link is not None
            except APIClientError as err:
                self.echo_debug(err)
                raise UploadError  # pylint: disable=W0707

    @backoff.on_exception(
        backoff.expo,
        (APIClientTimeout),
        max_tries=5,
        max_time=30,
    )
    def get_sample_sheet(self, next_link=None):
        """Get samples by gncv path."""
        return self.api_client.get_sample_sheet(
            self.destination,
            SampleAssignmentStatus.UNASSIGNED.value,
            next_link,
        )

    def output_list(self):
        """Output JSON of assigning samples to a project."""
        self.echo_debug("Outputting JSON.")
        if self.output == "-":
            self.echo_data(
                json.dumps(self.assigned_samples, indent=4, cls=CustomEncoder)
            )
        else:
            dirname = os.path.dirname(self.output)
            if dirname and not os.path.exists(dirname):
                os.makedirs(dirname)
            with open(self.output, "w") as json_file:
                json_file.write(
                    json.dumps(
                        self.assigned_samples, indent=4, cls=CustomEncoder
                    )
                )
            self.echo_info(
                "Assigned samples response outputted to {}".format(
                    self.output
                )
            )



"""Define common exceptions in Gencove CLI."""


class ValidationError(Exception):
    """CLI validation error - indicating command input is not valid."""

    def __init__(self, message=""):
        super().__init__(message)
        self.message = message



"""Utils for upload command."""
import csv
import os
import platform
from collections import defaultdict

from boto3.s3.transfer import TransferConfig

from botocore.exceptions import ClientError

from gencove.exceptions import ValidationError
from gencove.logger import echo_debug, echo_info
from gencove.utils import CHUNK_SIZE, get_progress_bar

from .constants import (
    FASTQ_EXTENSIONS,
    FastQ,
    PATH_TEMPLATE,
    PathTemplateParts,
    R_NOTATION_MAP,
)


def upload_file(
    s3_client, file_name, bucket, object_name=None, no_progress=False
):  # noqa: D413
    """Upload a file to an S3 bucket.

    Args:
        s3_client: Boto s3 client.
        file_name (str): File to upload.
        bucket (str): Bucket to upload to.
        object_name (str): S3 object name.
            If not specified then file_name is used

    Returns:
        True if file was uploaded, else False
    """
    # If S3 object_name was not specified, use file_name
    if object_name is None:
        object_name = file_name

    # Upload the file
    try:
        # Set desired multipart threshold value of 5GB
        config = TransferConfig(
            multipart_threshold=CHUNK_SIZE,
            multipart_chunksize=CHUNK_SIZE,
            use_threads=True,
            max_concurrency=10,
        )
        if not no_progress:
            progress_bar = get_progress_bar(
                os.path.getsize(file_name), "Uploading: "
            )
            progress_bar.start()
        s3_client.upload_file(
            file_name,
            bucket,
            object_name,
            Config=config,
            Callback=_progress_bar_update(progress_bar)
            if not no_progress
            else None,
        )
        if not no_progress:
            progress_bar.finish()
    except ClientError as err:
        echo_info("Failed to upload file {}: {}".format(file_name, err))
        return False
    return True


def upload_multi_file(
    s3_client,
    file_obj,
    bucket,
    object_name=None,  # pylint: disable=E0012,C0330
    no_progress=False,
):  # noqa: D413
    """Upload a file to an S3 bucket.

    Args:
        s3_client: Boto s3 client.
        file_obj (MultiFileReader): File-like object with read() and
            __iter__ methods
        bucket (str): Bucket to upload to.
        object_name (str): S3 object name.
            If not specified then file_name is used

    Returns:
        True if file was uploaded, else False
    """
    # If S3 object_name was not specified, use file_name
    if object_name is None:
        object_name = file_obj.name

    # Upload the file
    try:
        # Set desired multipart threshold value of 5GB
        config = TransferConfig(
            multipart_threshold=CHUNK_SIZE,
            multipart_chunksize=CHUNK_SIZE,
            use_threads=True,
            max_concurrency=10,
        )

        if not no_progress:
            progress_bar = get_progress_bar(
                file_obj.get_size(), "Uploading: "
            )
            progress_bar.start()
        s3_client.upload_fileobj(
            file_obj,
            bucket,
            object_name,
            Config=config,
            Callback=_progress_bar_update(progress_bar)
            if not no_progress
            else None,
        )
        if not no_progress:
            progress_bar.finish()
    except ClientError as err:
        echo_info("Failed to upload file {}: {}".format(file_obj.name, err))
        return False
    return True


def _progress_bar_update(pbar):  # noqa: D413
    """Update progress bar manually.

    Helper method for S3 Transfer,
    which needs a callback to update the progressbar.

    Args:
        pbar: progressbar.ProgressBar instance

    Returns:
        a function that in turn accepts chunk that is used to update the
        progressbar.
    """
    # noqa: D202
    def _update_pbar(chunk_uploaded_in_bytes):
        pbar.update(pbar.value + chunk_uploaded_in_bytes)

    return _update_pbar


def seek_files_to_upload(path, path_root=""):
    """Generate a list of valid fastq files."""
    for root, dirs, files in os.walk(path):
        files.sort()
        for file in files:
            file_path = os.path.join(path_root, root, file)

            if file_path.lower().endswith(FASTQ_EXTENSIONS):
                echo_debug("Found file to upload: {}".format(file_path))
                yield file_path

        dirs.sort()
        for folder in dirs:
            seek_files_to_upload(folder, root)


def get_get_upload_details_retry_predicate(resp):
    """Triggers retry if upload details came back without last status."""
    return not resp.last_status


def get_filename_from_path(full_path, source):
    """Cross OS get file name utility."""
    relpath = os.path.relpath(os.path.normpath(full_path), start=source)
    if platform.system() == "Windows":
        return relpath.replace("\\", "/")
    return relpath


def _validate_fastq(fastq):
    """Validate fastq object.

    Args:
        fastq (FastQ): fastq object
    Returns:
        None if all good
    Raises:
        ValidationError if fastq is not valid
    """
    if not fastq.path.lower().endswith(FASTQ_EXTENSIONS):
        echo_info("Unsupported file type found: {}".format(fastq.path))
        raise ValidationError(
            "Bad file extension in path: {}".format(fastq.path)
        )
    if "_" in fastq.client_id:
        echo_info("Underscore is not allowed in client id")
        raise ValidationError("Underscore is not allowed in client id")
    if not os.path.exists(fastq.path):
        echo_info("Path does not exist: {}".format(fastq.path))
        raise ValidationError("Could not find: {}".format(fastq.path))
    if fastq.r_notation not in R_NOTATION_MAP:
        echo_info("Wrong R notation: {}".format(fastq.r_notation))
        echo_info("Valid R notations are: {}".format(R_NOTATION_MAP.keys()))
        raise ValidationError("Wrong R notation: {}".format(fastq.r_notation))


def _validate_header(header):
    header_columns = header.values()
    for column in FastQ.__fields__:
        if column not in header_columns:
            raise ValidationError(
                "Unexpected CSV header. Expected: {}".format(
                    ", ".join(FastQ.__fields__)
                )
            )


def parse_fastqs_map_file(fastqs_map_path):
    """Parse fastq map file.

    Map file has to have following columns/headers:
        client_id, r_notation, path

    Example fastqs map file:
        client_id,r_notation,path
        sample1,R1,dir1/sample1_L001_R1.fastq.gz
        sample1,R1,dir1/sample1_L002_R1.fastq.gz
        sample2,R2,dir2/sample1_L001_R2.fastq.gz

    Args:
        fastqs_map_path (str): path to CSV file

    Returns:
        defaultdict: map of fastq file to samples
            {
                (<client_id>, <r_notation>): [path1, path2, ...],
            }
    """
    fastqs = defaultdict(list)
    with open(fastqs_map_path) as fastqs_file:
        reader = csv.DictReader(fastqs_file, fieldnames=FastQ.__fields__)
        # read headers row
        header = next(reader)
        _validate_header(header)
        for row in reader:
            fastq = FastQ(**row)
            _validate_fastq(fastq)
            fastqs[
                (fastq.client_id, R_NOTATION_MAP[fastq.r_notation])
            ].append(fastq.path)
    return fastqs


def get_gncv_path(client_id, r_notation):
    """Build gncv upload path.

    Args:
        client_id (str): sample client id
        r_notation (str): upload R1/2 notation

    Returns:
        str: gncv upload path
    """
    return PATH_TEMPLATE.format(
        **{
            PathTemplateParts.client_id.value: client_id,
            PathTemplateParts.r_notation.value: r_notation,
        }
    )



"""Test project's status merged VCF command."""

import io
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient, APIClientError, APIClientTimeout
from gencove.command.projects.cli import status_merged_vcf
from gencove.models import ProjectMergeVCFs


def test_status_merged_vcf__bad_project_id(mocker):
    """Test status merged file failure when non-uuid string is used as
    project id.
    """
    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_retrieve_merged_vcf = mocker.patch.object(
        APIClient,
        "retrieve_merged_vcf",
    )

    res = runner.invoke(
        status_merged_vcf,
        [
            "1111111",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_retrieve_merged_vcf.assert_not_called()
    assert "Project ID is not valid" in res.output


def test_status_merged_vcf__empty(mocker):
    """Test create merged file failure when project is responding empty.
    This can be due to no merged files or no permission required for the
    project or no project at all."""
    project_id = str(uuid4())
    mocked_response = {"detail": "Not found."}

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_retrieve_merged_vcf = mocker.patch.object(
        APIClient,
        "retrieve_merged_vcf",
        return_value=mocked_response,
        side_effect=APIClientError(message="", status_code=404),
    )

    res = runner.invoke(
        status_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_retrieve_merged_vcf.assert_called_once()
    message = (
        "Project {} does not exist or there are no "
        "running jobs associated with it.".format(project_id)
    )
    assert message in res.output


def test_status_merged_vcf__not_able(mocker):
    """Test status merged file failure when it cannot be done."""
    project_id = str(uuid4())
    mocked_response = [
        "You attempted to merge VCFs for a project that cannot do that"
    ]

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_retrieve_merged_vcf = mocker.patch.object(
        APIClient,
        "retrieve_merged_vcf",
        return_value=mocked_response,
        side_effect=APIClientError(
            message=mocked_response[0], status_code=400
        ),
    )

    res = runner.invoke(
        status_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_retrieve_merged_vcf.assert_called_once()
    assert (
        "You attempted to merge VCFs for a project that cannot do that"
        in res.output
    )


def test_status_merged_vcf__success_but_job_failed(mocker):
    """Test status merged VCF success, but the job is in the failed state."""
    project_id = str(uuid4())
    mocked_response = {
        "id": project_id,
        "created": "2020-09-14T08:59:00.480Z",
        "user": str(uuid4()),
        "last_status": {
            "id": str(uuid4()),
            "status": "failed",
            "note": "",
            "created": "2020-07-28T12:46:22.719862Z",
        },
        "up_to_date": False,
    }

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_retrieve_merged_vcf = mocker.patch.object(
        APIClient,
        "retrieve_merged_vcf",
        return_value=ProjectMergeVCFs(**mocked_response),
    )

    res = runner.invoke(
        status_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_retrieve_merged_vcf.assert_called_once()
    assert "WARNING: The job failed merging." in res.output


def test_status_merged_vcf__slow_response_retry(mocker):
    """Test status merged VCF slow response retry."""
    project_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_retrieve_merged_vcf = mocker.patch.object(
        APIClient,
        "retrieve_merged_vcf",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )

    res = runner.invoke(
        status_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    assert mocked_retrieve_merged_vcf.call_count == 5


def test_status_merged_vcf__success(mocker):
    """Test status merged VCF success."""
    project_id = str(uuid4())
    mocked_response = {
        "id": project_id,
        "created": "2020-09-14T08:59:00.480+00:00",
        "user": str(uuid4()),
        "last_status": {
            "id": str(uuid4()),
            "status": "success",
            "note": "",
            "created": "2020-07-28T12:46:22.719862+00:00",
        },
        "up_to_date": False,
    }

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_retrieve_merged_vcf = mocker.patch.object(
        APIClient,
        "retrieve_merged_vcf",
        return_value=ProjectMergeVCFs(**mocked_response),
    )

    res = runner.invoke(
        status_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_retrieve_merged_vcf.assert_called_once()
    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\t".join(
            [
                mocked_response["id"],
                mocked_response["last_status"]["created"],
                mocked_response["last_status"]["status"],
            ]
        )
    )
    assert output_line.getvalue() in res.output.encode()



"""Describe all constants in Gencove CLI."""
from enum import Enum, unique
from typing import Optional

from pydantic import BaseModel  # pylint: disable=no-name-in-module

HOST = "https://api.gencove.com"


@unique
class ApiEndpoints(Enum):
    """ApiEndpoints enum"""

    GET_JWT = "/api/v2/jwt-create/"
    REFRESH_JWT = "/api/v2/jwt-refresh/"
    UPLOAD_DETAILS = "/api/v2/uploads-post-data/"
    GET_UPLOAD_CREDENTIALS = "/api/v2/upload-credentials/"
    PROJECT_SAMPLES = "/api/v2/project-samples/{id}"
    SAMPLE_DETAILS = "/api/v2/samples/{id}"
    SAMPLE_QC_METRICS = "/api/v2/sample-quality-controls/{id}"
    SAMPLE_SHEET = "/api/v2/sample-sheet/"
    PROJECTS = "/api/v2/projects/"
    PIPELINE_CAPABILITES = "/api/v2/pipeline-capabilities/{id}"
    PROJECT_BATCH_TYPES = "/api/v2/project-batch-types/{id}"
    PROJECT_BATCHES = "/api/v2/project-batches/{id}"
    BATCHES = "/api/v2/batches/{id}"
    PROJECT_MERGE_VCFS = "/api/v2/project-merge-vcfs/{id}"
    SAMPLE_METADATA = "/api/v2/sample-metadata/{id}"
    PROJECT_RESTORE_SAMPLES = "/api/v2/project-restore-samples/{id}"
    BASESPACE_PROJECTS_IMPORT = "/api/v2/basespace-projects-import/"
    BASESPACE_PROJECTS_LIST = "/api/v2/basespace-projects/"
    BASESPACE_BIOSAMPLES_LIST = "/api/v2/basespace-project-biosamples/{id}"
    S3_URI_IMPORT = "/api/v2/s3-uri-import/"
    BASESPACE_PROJECTS_AUTOIMPORT = "/api/v2/basespace-projects-autoimport/"
    S3_URI_AUTOIMPORT = "/api/v2/s3-uri-autoimport/"


@unique
class SampleAssignmentStatus(Enum):
    """SampleAssignmentStatus enum"""

    ALL = "all"
    UNASSIGNED = "unassigned"
    ASSIGNED = "assigned"


@unique
class SampleSheetSortBy(Enum):
    """SampleSheetSortBy enum"""

    CREATED = "created"
    MODIFIED = "modified"


@unique
class SampleStatus(Enum):
    """SampleStatus enum"""

    COMPLETED = "completed"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    RUNNING = "running"
    ALL = "all"


@unique
class SampleArchiveStatus(Enum):
    """SampleArchiveStatus enum"""

    AVAILABLE = "available"
    ARCHIVED = "archived"
    RESTORE_REQUESTED = "restore_requested"
    ALL = "all"


@unique
class SampleSortBy(Enum):
    """SampleSortBy enum"""

    CREATED = "created"
    MODIFIED = "modified"
    STATUS = "status"
    CLIENT_ID = "client_id"
    ID = "id"


@unique
class SortOrder(Enum):
    """SortOrder enum"""

    ASC = "asc"
    DESC = "desc"


# pylint: disable=too-few-public-methods
class Credentials(BaseModel):
    """Credentials model"""

    email: str
    password: str
    api_key: str
    otp_token: Optional[str] = None


# pylint: disable=too-few-public-methods
class Optionals(BaseModel):
    """Optionals model"""

    host: Optional[str]


@unique
class DownloadTemplateParts(Enum):
    """DownloadTemplateParts enum"""

    CLIENT_ID = "client_id"
    GENCOVE_ID = "gencove_id"
    FILE_TYPE = "file_type"
    FILE_EXTENSION = "file_extension"
    DEFAULT_FILENAME = "default_filename"


DOWNLOAD_TEMPLATE = "{{{}}}/{{{}}}/{{{}}}".format(
    DownloadTemplateParts.CLIENT_ID.value,
    DownloadTemplateParts.GENCOVE_ID.value,
    DownloadTemplateParts.DEFAULT_FILENAME.value,
)

MAX_RETRY_TIME_SECONDS = 300  # 5 minutes
FASTQ_MAP_EXTENSION = ".fastq-map.csv"
UPLOAD_PREFIX = "gncv://"
ASSIGN_BATCH_SIZE = 200



"""Upload command exceptions"""


class UploadError(Exception):
    """Upload related error."""


class UploadNotFound(Exception):
    """Upload related error."""


class SampleSheetError(Exception):
    """Error to generate the sample sheet for uploads."""



"""Sample download file shell command definition."""
import sys

import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals
from gencove.logger import echo_error

from .main import DownloadFile


@click.command("download-file")
@click.argument("sample_id")
@click.argument("file_type")
@click.argument("destination")
@add_options(common_options)
@click.option(
    "--no-progress",
    is_flag=True,
    help="If specified, no progress bar is shown.",
)
# pylint: disable=too-many-arguments
def download_file(
    sample_id,
    file_type,
    destination,
    host,
    email,
    password,
    api_key,
    no_progress,
):  # noqa: D413,D301,D412 # pylint: disable=C0301
    """Download sample file metadata.

    SAMPLE_ID   specific sample for which to download the results

    FILE_TYPE   specific deliverable to download results for

    DESTINATION path/to/file

    Examples:

        Download sample:

            gencove samples download-file e6b45af7-07c5-4a6d-9f97-6e1efbf3e215 ancestry-json ancestry.json

        Download and print to stdout then compress using gzip:

            gencove samples download-file e6b45af7-07c5-4a6d-9f97-6e1efbf3e215 ancestry-json - | gzip > ancestry.json.gz
    \f

    Args:
        sample_id (str): specific sample for which
            to download the results.
        file_type (str): specific deliverable to download
            results for.
        destination (str): path/to/file.
        no_progress (bool, optional, default False): do not show progress
            bar.
    """  # noqa: E501
    if destination in ("-", "/dev/stdout"):
        DownloadFile(
            sample_id,
            file_type,
            sys.stdout.buffer,
            Credentials(email=email, password=password, api_key=api_key),
            Optionals(host=host),
            no_progress,
        ).run()
    else:
        try:
            with open(destination, "wb") as destination_file:
                DownloadFile(
                    sample_id,
                    file_type,
                    destination_file,
                    Credentials(
                        email=email, password=password, api_key=api_key
                    ),
                    Optionals(host=host),
                    no_progress,
                ).run()
        except IsADirectoryError:
            echo_error(
                "Please specify a file path (not directory path)"
                " for DESTINATION"
            )
            raise click.Abort()  # pylint: disable=raise-missing-from



"""AutoImport Biosamples from BaseSpace projects to a project subcommand."""

import json

from ....base import Command
from ....utils import is_valid_json, is_valid_uuid
from ..... import client
from .....exceptions import ValidationError


class BaseSpaceAutoImport(Command):
    """BaseSpace autoimport command executor."""

    def __init__(self, project_id, identifier, credentials, options):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.identifier = identifier
        self.metadata_json = options.metadata_json

    def initialize(self):
        """Initialize basespace-autoimport subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if is_valid_uuid(self.project_id) is False:
            raise ValidationError("Project ID is not valid. Exiting.")
        if self.metadata_json and is_valid_json(self.metadata_json) is False:
            raise ValidationError("Metadata JSON is not valid. Exiting.")

    # no retry for timeouts in order to avoid duplicate heavy operations on
    # the backend
    def execute(self):
        """Make a request to create a job to periodically import Biosamples
        from BaseSpace projects whose name contain the identifier to a
        Gencove project.
        """
        self.echo_debug(
            "Create AutoImport job of the "
            "identifier {} to a project {}".format(
                self.identifier, self.project_id
            )
        )

        try:
            # prepare metadata
            metadata = None
            if self.metadata_json is not None:
                metadata = json.loads(self.metadata_json)
                self.echo_info(
                    "Metadata will be assigned to the imported Biosamples."
                )
            autoimport_from_basespace = (
                self.api_client.autoimport_from_basespace(
                    project_id=self.project_id,
                    identifier=self.identifier,
                    metadata=metadata,
                )
            )
            self.echo_debug(autoimport_from_basespace)
            self.echo_info(
                "Request to create a periodic import job of BaseSpace "
                "projects accepted."
            )
        except client.APIClientError:
            self.echo_error(
                "There was an error creating a periodic import job of "
                "BaseSpace projects."
            )
            raise
